r"""
Helper functions for congruence subgroups

This file contains optimized Cython implementations of a few functions related
to the standard congruence subgroups `\Gamma_0, \Gamma_1, \Gamma_H`.  These
functions are for internal use by routines elsewhere in the Sage library.
"""

# ****************************************************************************
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from cysignals.memory cimport check_allocarray, sig_free

import random
from .congroup_gamma1 import Gamma1_constructor as Gamma1
from .congroup_gamma0 import Gamma0_constructor as Gamma0

cimport sage.rings.fast_arith
import sage.rings.fast_arith
cdef sage.rings.fast_arith.arith_int arith_int
arith_int = sage.rings.fast_arith.arith_int()
from sage.matrix.matrix_integer_dense cimport Matrix_integer_dense
from sage.modular.modsym.p1list import lift_to_sl2z
from sage.matrix.matrix_space import MatrixSpace
from sage.rings.integer_ring import ZZ
Mat2Z = MatrixSpace(ZZ, 2)

cdef Matrix_integer_dense genS, genT, genI
genS = Matrix_integer_dense(Mat2Z, [0, -1, 1, 0], True, True)
genT = Matrix_integer_dense(Mat2Z, [1, 1, 0, 1], True, True)
genI = Matrix_integer_dense(Mat2Z, [1, 0, 0, 1], True, True)


# This is the C version of a function formerly implemented in python in
# sage.modular.congroup.  It is orders of magnitude faster (e.g., 30
# times).  The key speedup is in replacing looping through the
# elements of the Python list R with looping through the elements of a
# C-array.

def degeneracy_coset_representatives_gamma0(int N, int M, int t):
    r"""
    Let `N` be a positive integer and `M` a divisor of `N`.  Let `t` be a
    divisor of `N/M`, and let `T` be the `2 \times 2` matrix `(1, 0; 0, t)`.
    This function returns representatives for the orbit set `\Gamma_0(N)
    \backslash T \Gamma_0(M)`, where `\Gamma_0(N)` acts on the left on `T
    \Gamma_0(M)`.

    INPUT:

    - ``N`` -- integer
    - ``M`` -- integer (divisor of `N`)
    - ``t`` -- integer (divisor of `N/M`)

    OUTPUT:

    list -- list of lists ``[a,b,c,d]``, where ``[a,b,c,d]`` should be viewed
    as a 2x2 matrix.

    This function is used for computation of degeneracy maps between
    spaces of modular symbols, hence its name.

    We use that `T^{-1} \cdot (a,b;c,d) \cdot T = (a,bt; c/t,d)`, that the
    group `T^{-1} \Gamma_0(N) T` is contained in `\Gamma_0(M)`, and that
    `\Gamma_0(N) T` is contained in `T \Gamma_0(M)`.

    ALGORITHM:

    1. Compute representatives for `\Gamma_0(N/t,t)` inside of `\Gamma_0(M)`:

      + COSET EQUIVALENCE: Two right cosets represented by `[a,b;c,d]` and
        `[a',b';c',d']` of `\Gamma_0(N/t,t)` in `\SL_2(\ZZ)` are equivalent if
        and only if `(a,b)=(a',b')` as points of `\mathbf{P}^1(\ZZ/t\ZZ)`,
        i.e., `ab' \cong ba' \pmod{t}`, and `(c,d) = (c',d')` as points of
        `\mathbf{P}^1(\ZZ/(N/t)\ZZ)`.

      + ALGORITHM to list all cosets:

        a) Compute the number of cosets.
        b) Compute a random element `x` of `\Gamma_0(M)`.
        c) Check if x is equivalent to anything generated so far; if not, add x
           to the list.
        d) Continue until the list is as long as the bound
           computed in step (a).

    2. There is a bijection between `\Gamma_0(N)\backslash T \Gamma_0(M)` and
       `\Gamma_0(N/t,t) \backslash \Gamma_0(M)` given by `T r \leftrightarrow
       r`. Consequently we obtain coset representatives for
       `\Gamma_0(N)\backslash T \Gamma_0(M)` by left multiplying by `T` each
       coset representative of `\Gamma_0(N/t,t) \backslash \Gamma_0(M)` found
       in step 1.

    EXAMPLES::

        sage: from sage.modular.arithgroup.all import degeneracy_coset_representatives_gamma0
        sage: len(degeneracy_coset_representatives_gamma0(13, 1, 1))
        14
        sage: len(degeneracy_coset_representatives_gamma0(13, 13, 1))
        1
        sage: len(degeneracy_coset_representatives_gamma0(13, 1, 13))
        14
    """
    if N % M != 0:
        raise ArithmeticError(f"M (={M}) must be a divisor of N (={N})")

    if (N // M) % t != 0:
        raise ArithmeticError(f"t (={t}) must be a divisor of N/M (={N//M})")

    cdef int n, i, j, k, aa, bb, cc, dd, g, Ndivt, halfmax, is_new
    cdef int* R

    # total number of coset representatives that we'll find
    n = Gamma0(N).index() / Gamma0(M).index()
    k = 0   # number found so far
    Ndivt = N // t
    R = <int*>check_allocarray(4 * n, sizeof(int))
    halfmax = 2*(n+10)
    while k < n:
        # try to find another coset representative.
        cc = M*random.randrange(-halfmax, halfmax+1)
        dd = random.randrange(-halfmax, halfmax+1)
        g = arith_int.c_xgcd_int(-cc, dd, &bb, &aa)
        if g == 0:
            continue
        cc = cc // g
        if cc % M != 0:
            continue
        dd = dd // g
        # Test if we've found a new coset representative.
        is_new = 1
        for i in range(k):
            j = 4*i
            if (R[j+1]*aa - R[j]*bb) % t == 0 and \
               (R[j+3]*cc - R[j+2]*dd) % Ndivt == 0:
                is_new = 0
                break
        # If our matrix is new add it to the list.
        if is_new:
            R[4*k] = aa
            R[4*k+1] = bb
            R[4*k+2] = cc
            R[4*k+3] = dd
            k = k + 1

    # Return the list left multiplied by T.
    S = []
    for i in range(k):
        j = 4*i
        S.append([R[j], R[j+1], R[j+2]*t, R[j+3]*t])
    sig_free(R)
    return S


def degeneracy_coset_representatives_gamma1(int N, int M, int t):
    r"""
    Let `N` be a positive integer and `M` a divisor of `N`.  Let `t` be a
    divisor of `N/M`, and let `T` be the `2 \times 2` matrix `(1,0; 0,t)`.
    This function returns representatives for the orbit set `\Gamma_1(N)
    \backslash T \Gamma_1(M)`, where `\Gamma_1(N)` acts on the left on `T
    \Gamma_1(M)`.

    INPUT:

    - ``N`` -- integer
    - ``M`` -- integer (divisor of `N`)
    - ``t`` -- integer (divisor of `N/M`)

    OUTPUT:

    list -- list of lists ``[a,b,c,d]``, where ``[a,b,c,d]`` should be viewed
    as a 2x2 matrix.

    This function is used for computation of degeneracy maps between
    spaces of modular symbols, hence its name.

    ALGORITHM:

    Everything is the same as for
    :func:`~degeneracy_coset_representatives_gamma0`, except for coset
    equivalence.   Here `\Gamma_1(N/t,t)` consists of matrices that are of the
    form `(1,*; 0,1) \bmod N/t` and `(1,0; *,1) \bmod t`.

    COSET EQUIVALENCE: Two right cosets represented by `[a,b;c,d]` and
    `[a',b';c',d']` of `\Gamma_1(N/t,t)` in `\SL_2(\ZZ)` are equivalent if
    and only if

    .. MATH::

        a \cong a' \pmod{t},
        b \cong b' \pmod{t},
        c \cong c' \pmod{N/t},
        d \cong d' \pmod{N/t}.

    EXAMPLES::

        sage: from sage.modular.arithgroup.all import degeneracy_coset_representatives_gamma1
        sage: len(degeneracy_coset_representatives_gamma1(13, 1, 1))
        168
        sage: len(degeneracy_coset_representatives_gamma1(13, 13, 1))
        1
        sage: len(degeneracy_coset_representatives_gamma1(13, 1, 13))
        168
    """
    if N % M != 0:
        raise ArithmeticError(f"M (={M}) must be a divisor of N (={N})")

    if (N // M) % t != 0:
        raise ArithmeticError(f"t (={t}) must be a divisor of N/M (={N//M})")

    cdef int d, g, i, j, k, n, aa, bb, cc, dd, Ndivt, halfmax, is_new
    cdef int* R

    # total number of coset representatives that we'll find
    n = Gamma1(N).index() / Gamma1(M).index()
    d = arith_int.c_gcd_int(t, N // t)
    n = n // d
    k = 0   # number found so far
    Ndivt = N // t
    R = <int*>check_allocarray(4 * n, sizeof(int))
    halfmax = 2*(n+10)
    while k < n:
        # try to find another coset representative.
        cc = M * random.randrange(-halfmax, halfmax + 1)
        dd = 1 + M * random.randrange(-halfmax, halfmax + 1)
        g = arith_int.c_xgcd_int(-cc, dd, &bb, &aa)
        if g == 0:
            continue
        cc = cc // g
        if cc % M != 0:
            continue
        dd = dd // g
        if M != 1 and dd % M != 1:
            continue
        # Test if we've found a new coset representative.
        is_new = 1
        for i in range(k):
            j = 4*i
            if (R[j] - aa) % t == 0 and \
               (R[j+1] - bb) % t == 0 and \
               (R[j+2] - cc) % Ndivt == 0 and \
               (R[j+3] - dd) % Ndivt == 0:
                is_new = 0
                break
        # If our matrix is new add it to the list.
        if is_new:
            if k > n:
                sig_free(R)
                raise RuntimeError("bug!!")
            R[4*k] = aa
            R[4*k+1] = bb
            R[4*k+2] = cc
            R[4*k+3] = dd
            k = k + 1

    # Return the list left multiplied by T.
    S = []
    for i in range(k):
        j = 4*i
        S.append([R[j], R[j+1], R[j+2]*t, R[j+3]*t])
    sig_free(R)
    return S


def generators_helper(coset_reps, level):
    r"""
    Helper function for generators of Gamma0, Gamma1 and GammaH.

    These are computed using coset representatives, via an "inverse
    Todd-Coxeter" algorithm, and generators for `\SL_2(\ZZ)`.

    ALGORITHM: Given coset representatives for a finite index subgroup `G` of
    `\SL_2(\ZZ)` we compute generators for `G` as follows.  Let `R` be a set of
    coset representatives for `G`.  Let `S, T \in \SL_2(\ZZ)` be defined by
    `(0,-1; 1,0)` and `(1,1,0,1)`, respectively.
    Define maps `s, t: R \to G` as follows. If `r \in R`, then there exists a
    unique `r' \in R` such that `GrS = Gr'`. Let `s(r) = rSr'^{-1}`. Likewise,
    there is a unique `r'` such that `GrT = Gr'` and we let `t(r) = rTr'^{-1}`.
    Note that `s(r)` and `t(r)` are in `G` for all `r`.  Then `G` is generated
    by `s(R)\cup t(R)`.

    There are more sophisticated algorithms using group actions on trees (and
    Farey symbols) that give smaller generating sets -- this code is now
    deprecated in favour of the newer implementation based on Farey symbols.

    EXAMPLES::

        sage: Gamma0(7).generators(algorithm='todd-coxeter') # indirect doctest
        [
        [1 1]  [-1  0]  [ 1 -1]  [1 0]  [1 1]  [-3 -1]  [-2 -1]  [-5 -1]
        [0 1], [ 0 -1], [ 0  1], [7 1], [0 1], [ 7  2], [ 7  3], [21  4],
        <BLANKLINE>
        [-4 -1]  [-1  0]  [ 1  0]
        [21  5], [ 7 -1], [-7  1]
        ]
    """
    cdef Matrix_integer_dense x, y, z, v, vSmod, vTmod

    crs = coset_reps.list()
    try:
        reps = [Matrix_integer_dense(Mat2Z, lift_to_sl2z(c, d, level),
                                     False, True) for c, d in crs]
    except (TypeError, ValueError, NotImplementedError):
        raise ArithmeticError("Error lifting to SL2Z: level=%s crs=%s" % (level, crs))
    ans = []
    cdef Py_ssize_t i
    for i in range(len(crs)):
        x = reps[i]
        v = Matrix_integer_dense(Mat2Z, [crs[i][0], crs[i][1], 0, 0],
                                 False, True)
        vSmod = (v*genS)
        vTmod = (v*genT)
        y_index = coset_reps.normalize(vSmod[0, 0], vSmod[0, 1])
        z_index = coset_reps.normalize(vTmod[0, 0], vTmod[0, 1])
        y_index = crs.index(y_index)
        z_index = crs.index(z_index)
        y = reps[y_index]
        z = reps[z_index]
        y = y.inverse_of_unit()
        z = z.inverse_of_unit()
        ans.append(x*genS*y)
        ans.append(x*genT*z)
    return [x for x in ans if x != genI]



r"""
Group of Tangent-Space Automorphism Fields

Given a differentiable manifold `U` and a differentiable map
`\Phi: U \rightarrow M` to a differentiable manifold `M` (possibly `U = M`
and `\Phi=\mathrm{Id}_M`), the *group of tangent-space automorphism fields*
associated with `U` and `\Phi` is the general linear group
`\mathrm{GL}(\mathfrak{X}(U,\Phi))` of the module `\mathfrak{X}(U,\Phi)` of
vector fields along `U` with values on `M\supset \Phi(U)` (see
:class:`~sage.manifolds.differentiable.vectorfield_module.VectorFieldModule`).
Note that `\mathfrak{X}(U, \Phi)` is a module over
`C^k(U)`, the algebra of differentiable scalar fields on `U`.
Elements of `\mathrm{GL}(\mathfrak{X}(U, \Phi))` are fields along `U`
of automorphisms of tangent spaces to `M`.

Two classes implement `\mathrm{GL}(\mathfrak{X}(U, \Phi))` depending
whether `M` is parallelizable or not:
:class:`AutomorphismFieldParalGroup` and :class:`AutomorphismFieldGroup`.

AUTHORS:

- Eric Gourgoulhon (2015): initial version
- Travis Scrimshaw (2016): review tweaks
- Michael Jung (2019): improve treatment of the identity element

REFERENCES:

- Chap. 15 of [God1968]_
"""

#******************************************************************************
#       Copyright (C) 2015 Eric Gourgoulhon <eric.gourgoulhon@obspm.fr>
#       Copyright (C) 2016 Travis Scrimshaw <tscrimsh@umn.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  http://www.gnu.org/licenses/
#******************************************************************************

from sage.categories.groups import Groups
from sage.manifolds.differentiable.automorphismfield import (
    AutomorphismField,
    AutomorphismFieldParal,
)
from sage.manifolds.differentiable.vectorfield_module import (
    VectorFieldFreeModule,
    VectorFieldModule,
)
from sage.misc.cachefunc import cached_method
from sage.structure.parent import Parent
from sage.structure.unique_representation import UniqueRepresentation
from sage.tensor.modules.free_module_linear_group import FreeModuleLinearGroup


class AutomorphismFieldGroup(UniqueRepresentation, Parent):
    r"""
    General linear group of the module of vector fields along a differentiable
    manifold `U` with values on a differentiable manifold `M`.

    Given a differentiable manifold `U` and a differentiable map
    `\Phi: U \rightarrow M` to a differentiable manifold `M` (possibly `U = M`
    and `\Phi = \mathrm{Id}_M`), the *group of tangent-space automorphism
    fields* associated with `U` and `\Phi` is the general linear group
    `\mathrm{GL}(\mathfrak{X}(U,\Phi))` of the module `\mathfrak{X}(U,\Phi)` of
    vector fields along `U` with values on `M \supset \Phi(U)` (see
    :class:`~sage.manifolds.differentiable.vectorfield_module.VectorFieldModule`).
    Note that `\mathfrak{X}(U,\Phi)` is a module over
    `C^k(U)`, the algebra of differentiable scalar fields on `U`.
    Elements of `\mathrm{GL}(\mathfrak{X}(U,\Phi))` are fields along `U` of
    automorphisms of tangent spaces to `M`.

    .. NOTE::

        If `M` is parallelizable, then :class:`AutomorphismFieldParalGroup`
        *must* be used instead.

    INPUT:

    - ``vector_field_module`` --
      :class:`~sage.manifolds.differentiable.vectorfield_module.VectorFieldModule`;
      module `\mathfrak{X}(U,\Phi)` of vector fields along `U` with values on `M`

    EXAMPLES:

    Group of tangent-space automorphism fields of the 2-sphere::

        sage: M = Manifold(2, 'M') # the 2-dimensional sphere S^2
        sage: U = M.open_subset('U') # complement of the North pole
        sage: c_xy.<x,y> = U.chart() # stereographic coordinates from the North pole
        sage: V = M.open_subset('V') # complement of the South pole
        sage: c_uv.<u,v> = V.chart() # stereographic coordinates from the South pole
        sage: M.declare_union(U,V)   # S^2 is the union of U and V
        sage: xy_to_uv = c_xy.transition_map(c_uv, (x/(x^2+y^2), y/(x^2+y^2)),
        ....:             intersection_name='W',
        ....:             restrictions1= x^2+y^2!=0, restrictions2= u^2+v^2!=0)
        sage: uv_to_xy = xy_to_uv.inverse()
        sage: G = M.automorphism_field_group() ; G
        General linear group of the Module X(M) of vector fields on the
         2-dimensional differentiable manifold M

    ``G`` is the general linear group of the vector field module
    `\mathfrak{X}(M)`::

        sage: XM = M.vector_field_module() ; XM
        Module X(M) of vector fields on the 2-dimensional differentiable
         manifold M
        sage: G is XM.general_linear_group()
        True

    ``G`` is a non-abelian group::

        sage: G.category()
        Category of groups
        sage: G in Groups()
        True
        sage: G in CommutativeAdditiveGroups()
        False

    The elements of ``G`` are tangent-space automorphisms::

        sage: a = G.an_element(); a
        Field of tangent-space automorphisms on the 2-dimensional
         differentiable manifold M
        sage: a.parent() is G
        True
        sage: a.restrict(U).display()
        2 ∂/∂x⊗dx + 2 ∂/∂y⊗dy
        sage: a.restrict(V).display()
        2 ∂/∂u⊗du + 2 ∂/∂v⊗dv

    The identity element of the group ``G``::

        sage: e = G.one() ; e
        Field of tangent-space identity maps on the 2-dimensional
         differentiable manifold M
        sage: eU = U.default_frame() ; eU
        Coordinate frame (U, (∂/∂x,∂/∂y))
        sage: eV = V.default_frame() ; eV
        Coordinate frame (V, (∂/∂u,∂/∂v))
        sage: e.display(eU)
        Id = ∂/∂x⊗dx + ∂/∂y⊗dy
        sage: e.display(eV)
        Id = ∂/∂u⊗du + ∂/∂v⊗dv
    """

    Element = AutomorphismField

    def __init__(self, vector_field_module):
        r"""
        See :class:`AutomorphismfieldGroup` for documentation and examples.

        TESTS::

            sage: M = Manifold(2, 'M')
            sage: U = M.open_subset('U') ; V = M.open_subset('V')
            sage: M.declare_union(U,V)   # M is the union of U and V
            sage: c_xy.<x,y> = U.chart() ; c_uv.<u,v> = V.chart()
            sage: transf = c_xy.transition_map(c_uv, (x+y, x-y),
            ....:  intersection_name='W', restrictions1= x>0,
            ....:  restrictions2= u+v>0)
            sage: inv = transf.inverse()
            sage: from sage.manifolds.differentiable.automorphismfield_group \
            ....:                                 import AutomorphismFieldGroup
            sage: G = AutomorphismFieldGroup(M.vector_field_module()) ; G
            General linear group of the Module X(M) of vector fields on the
             2-dimensional differentiable manifold M
            sage: TestSuite(G).run(skip='_test_elements')

        ``_test_elements`` does not pass due to the failure
        of ``_test_pickling`` in
        :class:`sage.manifolds.differentiable.tensorfield.TensorField`.
        """
        if not isinstance(vector_field_module, VectorFieldModule):
            raise TypeError("{} is not a module of vector fields".format(
                            vector_field_module))
        Parent.__init__(self, category=Groups())
        self._vmodule = vector_field_module

    #### Parent methods ####

    def _element_constructor_(self, comp=[], frame=None, name=None,
                              latex_name=None):
        r"""
        Construct a field of tangent-space automorphisms.

        OUTPUT: :class:`~sage.manifolds.differentiable.automorphismfield.AutomorphismField`

        TESTS::

            sage: M = Manifold(2, 'M')
            sage: U = M.open_subset('U'); V = M.open_subset('V')
            sage: M.declare_union(U,V)   # M is the union of U and V
            sage: c_xy.<x,y> = U.chart(); c_uv.<u,v> = V.chart()
            sage: transf = c_xy.transition_map(c_uv, (x+y, x-y),
            ....:  intersection_name='W', restrictions1= x>0,
            ....:  restrictions2= u+v>0)
            sage: inv = transf.inverse()
            sage: G = M.automorphism_field_group()
            sage: a = G(1); a
            Field of tangent-space identity maps on the 2-dimensional
             differentiable manifold M
            sage: a = G([[1+x^2, 0], [0, 1+y^2]], frame=c_xy.frame(), name='a'); a
            Field of tangent-space automorphisms a on the 2-dimensional
             differentiable manifold M
            sage: a.display(c_xy.frame())
            a = (x^2 + 1) ∂/∂x⊗dx + (y^2 + 1) ∂/∂y⊗dy
        """
        if hasattr(comp, 'is_trivial_zero'):
            if (comp - 1).is_trivial_zero():
                return self.one()
        elif comp == 1:
            return self.one()
        if not isinstance(comp, (list, tuple)):
            raise TypeError("cannot convert the {} ".format(comp) +
                            "to an element of {}".format(self))
        # standard construction
        resu = self.element_class(self._vmodule, name=name,
                                  latex_name=latex_name)
        if comp:
            resu.set_comp(frame)[:] = comp
        return resu

    def _an_element_(self):
        r"""
        Construct some specific field of tangent-space automorphisms.

        OUTPUT: :class:`~sage.manifolds.differentiable.automorphismfield.AutomorphismField`

        TESTS::

            sage: M = Manifold(2, 'M')
            sage: U = M.open_subset('U') ; V = M.open_subset('V')
            sage: M.declare_union(U,V)   # M is the union of U and V
            sage: c_xy.<x,y> = U.chart() ; c_uv.<u,v> = V.chart()
            sage: transf = c_xy.transition_map(c_uv, (x+y, x-y),
            ....:  intersection_name='W', restrictions1= x>0,
            ....:  restrictions2= u+v>0)
            sage: inv = transf.inverse()
            sage: G = M.automorphism_field_group()
            sage: a = G.an_element() ; a
            Field of tangent-space automorphisms on the 2-dimensional
             differentiable manifold M
            sage: a.restrict(U).display()
            2 ∂/∂x⊗dx + 2 ∂/∂y⊗dy
            sage: a.restrict(V).display()
            2 ∂/∂u⊗du + 2 ∂/∂v⊗dv
            sage: a == G.an_element()  # indirect doctest
            True
        """
        resu = self.element_class(self._vmodule)
        for dom in resu.domain().subsets():
            if dom.is_manifestly_parallelizable():
                fmodule = dom.vector_field_module()
                idm = fmodule.identity_map()
                rst = fmodule.automorphism()
                for frame, comp in idm._components.items():
                    rst._components[frame] = 2 * comp
                resu._restrictions[dom] = rst
        return resu

    #### End of parent methods ####

    #### Monoid methods ####

    @cached_method
    def one(self):
        r"""
        Return identity element of ``self``.

        The group identity element is the field of tangent-space identity maps.

        OUTPUT:

        - :class:`~sage.manifolds.differentiable.automorphismfield.AutomorphismField`
          representing the identity element

        EXAMPLES:

        Identity element of the group of tangent-space automorphism fields of
        the 2-sphere::

            sage: M = Manifold(2, 'M') # the 2-dimensional sphere S^2
            sage: U = M.open_subset('U') # complement of the North pole
            sage: c_xy.<x,y> = U.chart() # stereographic coordinates from the North pole
            sage: V = M.open_subset('V') # complement of the South pole
            sage: c_uv.<u,v> = V.chart() # stereographic coordinates from the South pole
            sage: M.declare_union(U,V)   # S^2 is the union of U and V
            sage: xy_to_uv = c_xy.transition_map(c_uv, (x/(x^2+y^2), y/(x^2+y^2)),
            ....:                                intersection_name='W', restrictions1= x^2+y^2!=0,
            ....:                                restrictions2= u^2+v^2!=0)
            sage: uv_to_xy = xy_to_uv.inverse()
            sage: G = M.automorphism_field_group()
            sage: G.one()
            Field of tangent-space identity maps on the 2-dimensional differentiable manifold M
            sage: G.one().restrict(U)[:]
            [1 0]
            [0 1]
            sage: G.one().restrict(V)[:]
            [1 0]
            [0 1]
        """
        # Specific initializations for the field of identity maps:
        resu = self._element_constructor_(name='Id', latex_name=r'\mathrm{Id}')
        resu._inverse = resu
        for dom in resu._domain._subsets:
            if dom.is_manifestly_parallelizable():
                fmodule = dom.vector_field_module()
                resu._restrictions[dom] = fmodule.identity_map(name='Id',
                                                     latex_name=r'\mathrm{Id}')
        resu._is_identity = True
        resu.set_immutable()
        return resu

    #### End of monoid methods ####

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        TESTS::

            sage: M = Manifold(2, 'M')
            sage: G = M.automorphism_field_group()
            sage: G._repr_()
            'General linear group of the Module X(M) of vector fields on the
             2-dimensional differentiable manifold M'
            sage: repr(G)  # indirect doctest
            'General linear group of the Module X(M) of vector fields on the
             2-dimensional differentiable manifold M'
            sage: G  # indirect doctest
            General linear group of the Module X(M) of vector fields on the
             2-dimensional differentiable manifold M
        """
        return "General linear group of the {}".format(self._vmodule)

    def _latex_(self):
        r"""
        Return a LaTeX representation of ``self``.

        TESTS::

            sage: M = Manifold(2, 'M')
            sage: G = M.automorphism_field_group()
            sage: G._latex_()
            \mathrm{GL}\left( \mathfrak{X}\left(M\right) \right)
            sage: latex(G)  # indirect doctest
            \mathrm{GL}\left( \mathfrak{X}\left(M\right) \right)
        """
        from sage.misc.latex import latex
        return r"\mathrm{GL}\left(" + latex(self._vmodule) + r"\right)"

    def base_module(self):
        r"""
        Return the vector-field module of which ``self`` is the general
        linear group.

        OUTPUT: :class:`~sage.manifolds.differentiable.vectorfield_module.VectorFieldModule`

        EXAMPLES:

        Base module of the group of tangent-space automorphism fields of
        the 2-sphere::

            sage: M = Manifold(2, 'M') # the 2-dimensional sphere S^2
            sage: U = M.open_subset('U') # complement of the North pole
            sage: c_xy.<x,y> = U.chart() # stereographic coordinates from the North pole
            sage: V = M.open_subset('V') # complement of the South pole
            sage: c_uv.<u,v> = V.chart() # stereographic coordinates from the South pole
            sage: M.declare_union(U,V)   # S^2 is the union of U and V
            sage: xy_to_uv = c_xy.transition_map(c_uv, (x/(x^2+y^2), y/(x^2+y^2)),
            ....:                                intersection_name='W', restrictions1= x^2+y^2!=0,
            ....:                                restrictions2= u^2+v^2!=0)
            sage: uv_to_xy = xy_to_uv.inverse()
            sage: G = M.automorphism_field_group()
            sage: G.base_module()
            Module X(M) of vector fields on the 2-dimensional differentiable
             manifold M
            sage: G.base_module() is M.vector_field_module()
            True
        """
        return self._vmodule


#******************************************************************************

class AutomorphismFieldParalGroup(FreeModuleLinearGroup):
    r"""
    General linear group of the module of vector fields along a differentiable
    manifold `U` with values on a parallelizable manifold `M`.

    Given a differentiable manifold `U` and a differentiable map
    `\Phi: U \rightarrow M` to a parallelizable  manifold `M` (possibly `U = M`
    and `\Phi = \mathrm{Id}_M`), the *group of tangent-space automorphism
    fields* associated with `U` and `\Phi` is the general linear group
    `\mathrm{GL}(\mathfrak{X}(U, \Phi))` of the module `\mathfrak{X}(U, \Phi)`
    of vector fields along `U` with values on `M \supset \Phi(U)` (see
    :class:`~sage.manifolds.differentiable.vectorfield_module.VectorFieldFreeModule`).
    Note that `\mathfrak{X}(U, \Phi)` is a free module over `C^k(U)`,
    the algebra of differentiable scalar fields on `U`.
    Elements of `\mathrm{GL}(\mathfrak{X}(U, \Phi))` are fields along `U` of
    automorphisms of tangent spaces to `M`.

    .. NOTE::

        If `M` is not parallelizable, the class
        :class:`AutomorphismFieldGroup` must be used instead.

    INPUT:

    - ``vector_field_module`` --
      :class:`~sage.manifolds.differentiable.vectorfield_module.VectorFieldFreeModule`;
      free module `\mathfrak{X}(U,\Phi)` of vector fields along `U`
      with values on `M`

    EXAMPLES:

    Group of tangent-space automorphism fields of a 2-dimensional
    parallelizable manifold::

        sage: M = Manifold(2, 'M')
        sage: X.<x,y> = M.chart()
        sage: XM = M.vector_field_module() ; XM
        Free module X(M) of vector fields on the 2-dimensional differentiable
         manifold M
        sage: G = M.automorphism_field_group(); G
        General linear group of the Free module X(M) of vector fields on the
         2-dimensional differentiable manifold M
        sage: latex(G)
        \mathrm{GL}\left( \mathfrak{X}\left(M\right) \right)

    ``G`` is nothing but the general linear group of the module
    `\mathfrak{X}(M)`::

        sage: G is XM.general_linear_group()
        True

    ``G`` is a group::

        sage: G.category()
        Category of groups
        sage: G in Groups()
        True

    It is not an abelian group::

        sage: G in CommutativeAdditiveGroups()
        False

    The elements of ``G`` are tangent-space automorphisms::

        sage: G.Element
        <class 'sage.manifolds.differentiable.automorphismfield.AutomorphismFieldParal'>
        sage: a = G.an_element() ; a
        Field of tangent-space automorphisms on the 2-dimensional
         differentiable manifold M
        sage: a.parent() is G
        True

    As automorphisms of `\mathfrak{X}(M)`, the elements of ``G`` map a vector
    field to a vector field::

        sage: v = XM.an_element() ; v
        Vector field on the 2-dimensional differentiable manifold M
        sage: v.display()
        2 ∂/∂x + 2 ∂/∂y
        sage: a(v)
        Vector field on the 2-dimensional differentiable manifold M
        sage: a(v).display()
        2 ∂/∂x - 2 ∂/∂y

    Indeed the matrix of ``a`` with respect to the frame
    `(\partial_x, \partial_y)` is::

        sage: a[X.frame(),:]
        [ 1  0]
        [ 0 -1]

    The elements of ``G`` can also be considered as tensor fields of
    type `(1,1)`::

        sage: a.tensor_type()
        (1, 1)
        sage: a.tensor_rank()
        2
        sage: a.domain()
        2-dimensional differentiable manifold M
        sage: a.display()
        ∂/∂x⊗dx - ∂/∂y⊗dy

    The identity element of the group ``G`` is::

        sage: id = G.one() ; id
        Field of tangent-space identity maps on the 2-dimensional
         differentiable manifold M
        sage: id*a == a
        True
        sage: a*id == a
        True
        sage: a*a^(-1) == id
        True
        sage: a^(-1)*a == id
        True

    Construction of an element by providing its components with
    respect to the manifold's default frame (frame associated to
    the coordinates `(x,y)`)::

        sage: b = G([[1+x^2,0], [0,1+y^2]]) ; b
        Field of tangent-space automorphisms on the 2-dimensional
         differentiable manifold M
        sage: b.display()
        (x^2 + 1) ∂/∂x⊗dx + (y^2 + 1) ∂/∂y⊗dy
        sage: (~b).display()  # the inverse automorphism
        1/(x^2 + 1) ∂/∂x⊗dx + 1/(y^2 + 1) ∂/∂y⊗dy

    We check the group law on these elements::

        sage: (a*b)^(-1) == b^(-1) * a^(-1)
        True

    Invertible tensor fields of type `(1,1)` can be converted to
    elements of ``G``::

        sage: t = M.tensor_field(1, 1, name='t')
        sage: t[:] = [[1+exp(y), x*y], [0, 1+x^2]]
        sage: t1 = G(t) ; t1
        Field of tangent-space automorphisms t on the 2-dimensional
         differentiable manifold M
        sage: t1 in G
        True
        sage: t1.display()
        t = (e^y + 1) ∂/∂x⊗dx + x*y ∂/∂x⊗dy + (x^2 + 1) ∂/∂y⊗dy
        sage: t1^(-1)
        Field of tangent-space automorphisms t^(-1) on the 2-dimensional
         differentiable manifold M
        sage: (t1^(-1)).display()
        t^(-1) = 1/(e^y + 1) ∂/∂x⊗dx - x*y/(x^2*(e^y + 1) + e^y + 1) ∂/∂x⊗dy + 1/(x^2 + 1) ∂/∂y⊗dy

    Since any automorphism field can be considered as a tensor field of
    type-`(1,1)` on ``M``, there is a coercion map from ``G`` to the
    module `T^{(1,1)}(M)` of type-`(1,1)` tensor fields::

        sage: T11 = M.tensor_field_module((1,1)) ; T11
        Free module T^(1,1)(M) of type-(1,1) tensors fields on the
         2-dimensional differentiable manifold M
        sage: T11.has_coerce_map_from(G)
        True

    An explicit call of this coercion map is::

        sage: tt = T11(t1) ; tt
        Tensor field t of type (1,1) on the 2-dimensional differentiable
         manifold M
        sage: tt == t
        True

    An implicit call of the coercion map is performed to subtract an
    element of ``G`` from an element of `T^{(1,1)}(M)`::

        sage: s = t - t1 ; s
        Tensor field t-t of type (1,1) on
         the 2-dimensional differentiable manifold M
        sage: s.parent() is T11
        True
        sage: s.display()
        t-t = 0

    as well as for the reverse operation::

        sage: s = t1 - t ; s
        Tensor field t-t of type (1,1) on the 2-dimensional differentiable
         manifold M
        sage: s.display()
        t-t = 0

    TESTS::

        sage: TestSuite(G).run()
    """

    Element = AutomorphismFieldParal

    def __init__(self, vector_field_module):
        r"""
        See :class:`AutomorphismfieldParalGroup` for documentation and
        examples.

        TESTS::

            sage: M = Manifold(2, 'M') ; M
            2-dimensional differentiable manifold M
            sage: X.<x,y> = M.chart()  # makes M parallelizable
            sage: G = M.automorphism_field_group(); G
            General linear group of the Free module X(M) of vector fields on
             the 2-dimensional differentiable manifold M
            sage: TestSuite(G).run()
        """
        if not isinstance(vector_field_module, VectorFieldFreeModule):
            raise TypeError("{} is not a free module of vector fields".format(
                            vector_field_module))
        FreeModuleLinearGroup.__init__(self, vector_field_module)



cdef inline type element_class_zero = pAdicRelaxedElement_zero
cdef inline type element_class_one = pAdicRelaxedElement_one
cdef inline type element_class_bound = pAdicRelaxedElement_bound
cdef inline type element_class_value = pAdicRelaxedElement_value
cdef inline type element_class_random = pAdicRelaxedElement_random
cdef inline type element_class_slice = pAdicRelaxedElement_slice
cdef inline type element_class_add = pAdicRelaxedElement_add
cdef inline type element_class_sub = pAdicRelaxedElement_sub
cdef inline type element_class_mul = pAdicRelaxedElement_mul
cdef inline type element_class_muldigit = pAdicRelaxedElement_muldigit
cdef inline type element_class_div = pAdicRelaxedElement_div
cdef inline type element_class_sqrt = pAdicRelaxedElement_sqrt
cdef inline type element_class_teichmuller = pAdicRelaxedElement_teichmuller
cdef inline type element_class_unknown = pAdicRelaxedElement_unknown

include "sage/libs/linkages/padics/relaxed/flint.pxi"
include "relaxed_template.pxi"



r"""
Elements of Drinfeld modular forms rings

This module defines the elements of the class
:class:`~sage.modular.drinfeld_modform.ring.DrinfeldModularForms`.

AUTHORS:

- David Ayotte (2022): initial version
"""

# ****************************************************************************
#       Copyright (C) 2022 DAVID AYOTTE <da.ayotte@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.structure.element import ModuleElement
from sage.structure.richcmp import richcmp, op_NE, op_EQ

from sage.rings.integer_ring import ZZ
from sage.rings.polynomial.multi_polynomial import MPolynomial


class DrinfeldModularFormsElement(ModuleElement):
    r"""
    Element class of rings of Drinfeld modular forms.

    Recall that a *graded Drinfeld form* is a sum of Drinfeld modular
    forms having potentially different weights:

    .. MATH::

        F = f_{k_1} + f_{k_2} + \cdots + f_{k_n}

    where `f_{k_i}` is a Drinfeld modular form of weight `k_i`. We also
    say that `f_{k_i}` is an *homogeneous component of weight* `k_i`. If
    `n=1`, then we say that `F` is *homogeneous of weight* `k_1`.

    EXAMPLES: use the ``inject_variable`` method of the parent to
    quickly assign variables names to the generators::

        sage: A = GF(3)['T']
        sage: K.<T> = Frac(A)
        sage: M = DrinfeldModularForms(K, 2)
        sage: M.inject_variables()
        Defining g1, g2
        sage: g1 in M
        True
        sage: g2.parent()
        Ring of Drinfeld modular forms of rank 2 over Fraction Field of Univariate Polynomial Ring in T over Finite Field of size 3

    Next, via algebraic combination of the generator, we may create any
    element of the ring::

        sage: F = g1*g2 + g2
        sage: F
        g1*g2 + g2
        sage: F.is_homogeneous()
        False
        sage: F.homogeneous_components()
        {8: g2, 10: g1*g2}

    If the created form is homogeneous, we can ask for its weight in
    which case it will be a Drinfeld modular form::

        sage: H = g1^4*g2^9 + T*g1^8*g2^8 + (T^2 - 1)*g1^28*g2^3
        sage: H.is_homogeneous()
        True
        sage: H.weight()
        80

    You can also construct an element by simply passing a multivariate
    polynomial to the parent::

        sage: f1, f2 = polygens(K, 2, 'f1, f2')
        sage: M(f1)
        g1
        sage: M(f2)
        g2
        sage: M(T*f1 + f2^3 + T^2 + 1)
        g2^3 + T*g1 + (T^2 + 1)

    .. NOTE::

        This class should not be directly instantiated, instead create
        an instance of the parent
        :class:`~sage.modular.drinfeld_modform.ring.DrinfeldModularForms`
        and access its elements using the relevant methods.
    """
    def __init__(self, parent, polynomial):
        if not isinstance(polynomial, MPolynomial):
            raise TypeError("input must be a multivariate polynomial")
        if not parent.base_ring().has_coerce_map_from(polynomial.base_ring()):
            raise ValueError("unable to coerce base ring of the given "
                             "polynomial into Drinfeld modular form ring")
        poly = parent._poly_ring(polynomial)
        self._polynomial = poly

        super().__init__(parent)

    def _repr_(self):
        r"""
        Return the string representation of ``self``.

        TESTS::

            sage: A = GF(3)['T']; K = Frac(A)
            sage: M = DrinfeldModularForms(K, 2)
            sage: (M.0)._repr_()
            'g1'
            sage: M.0 + M.1
            g2 + g1
        """
        return str(self._polynomial)

    def _add_(self, other):
        r"""
        Return the addition of ``self`` with ``other``.

        TESTS::

            sage: A = GF(3)['T']; K = Frac(A)
            sage: M = DrinfeldModularForms(K, 2)
            sage: M.0 + M.1  # indirect doctest
            g2 + g1
        """
        return self.__class__(self.parent(), self._polynomial + other._polynomial)

    def _mul_(self, other):
        r"""
        Return the multiplication of ``self`` with ``other``.

        TESTS::

            sage: A = GF(3)['T']; K = Frac(A)
            sage: M = DrinfeldModularForms(K, 2)
            sage: M.0*M.1  # indirect doctest
            g1*g2
            sage: M.0*(M.0 + M.1)
            g1*g2 + g1^2
            sage: (M.0 + M.1)*M.0
            g1*g2 + g1^2
        """
        return self.__class__(self.parent(), self._polynomial*other._polynomial)

    def _lmul_(self, c):
        r"""
        Return the scalar multiplication of ``self`` by `c`.

        TESTS::

            sage: A = GF(3)['T']; K = Frac(A); T = K.gen()
            sage: M = DrinfeldModularForms(K, 2)
            sage: (T^2 + T + 2) * M.0  # indirect doctest
            (T^2 + T - 1)*g1
            sage: M.1 * (T^5 + T^2)
            (T^5 + T^2)*g2
            sage: 0 * M.1
            0
            sage: M.0 * 0
            0
        """
        return self.__class__(self.parent(), c*self._polynomial)

    def __neg__(self):
        r"""
        Return the negation of ``self``.

        TESTS::

            sage: A = GF(3)['T']; K = Frac(A)
            sage: M = DrinfeldModularForms(K, 2)
            sage: -M.0  # indirect doctest
            -g1
        """
        return self.__class__(self.parent(), -self._polynomial)

    def __bool__(self):
        r"""
        Return ``True`` whether ``self`` is nonzero.

        TESTS::

            sage: A = GF(3)['T']; K = Frac(A)
            sage: M = DrinfeldModularForms(K, 2)
            sage: bool(M.0)
            True
        """
        return bool(self._polynomial)

    def _latex_(self):
        r"""
        Return the LaTeX expression of ``self``.

        TESTS::

            sage: A = GF(3)['T']; K.<T> = Frac(A)
            sage: M = DrinfeldModularForms(K, 2)
            sage: M.inject_variables()
            Defining g1, g2
            sage: latex(g1)
            g_{1}
            sage: latex(g2)
            g_{2}
            sage: latex(1/T*g1^5 + g2*g1)
            \frac{1}{T} g_{1}^{5} + g_{1} g_{2}
        """
        return self._polynomial._latex_()

    def _richcmp_(self, other, op):
        r"""
        Return the comparison of ``self`` with ``other``.

        TESTS::

            sage: A = GF(3)['T']; K = Frac(A)
            sage: M = DrinfeldModularForms(K, 2)
            sage: M.0 == M.1
            False
            sage: M.0 != M.1
            True
            sage: M.0 == M.0
            True
            sage: M.0 < M.1
            Traceback (most recent call last):
            ...
            TypeError: '<' not supported between instances of 'DrinfeldModularForms_with_category.element_class' and 'DrinfeldModularForms_with_category.element_class'
        """
        if op != op_EQ and op != op_NE:
            return NotImplemented
        return richcmp(self._polynomial, other._polynomial, op)

    def rank(self):
        r"""
        Return the rank of this graded Drinfeld form.

        Note that the rank is independent of the chosen form and depends
        only on the parent.

        EXAMPLES::

            sage: A = GF(3)['T']; K = Frac(A)
            sage: M2 = DrinfeldModularForms(K, 2)
            sage: (M2.0).rank()
            2
            sage: M5 = DrinfeldModularForms(K, 5)
            sage: (M5.0 + M5.3).rank()
            5
        """
        return self.parent()._rank

    def is_one(self) -> bool:
        r"""
        Return ``True`` whether this graded Drinfeld form is the
        multiplicative identity.

        EXAMPLES::

            sage: A = GF(3)['T']; K = Frac(A)
            sage: M = DrinfeldModularForms(K, 2)
            sage: u = M.one()
            sage: u.is_one()
            True
            sage: (M.0).is_one()
            False
        """
        return self._polynomial.is_one()

    def is_zero(self) -> bool:
        r"""
        Return ``True`` whether this graded Drinfeld form is the
        additive identity.

        EXAMPLES::

            sage: A = GF(3)['T']; K = Frac(A)
            sage: M = DrinfeldModularForms(K, 2)
            sage: z = M.zero()
            sage: z.is_zero()
            True
            sage: f = M.0
            sage: f.is_zero()
            False
            sage: (f - f).is_zero()
            True
            sage: (0 * M.0).is_zero()
            True
        """
        return not bool(self)

    def is_homogeneous(self) -> bool:
        r"""
        Return whether the graded form is homogeneous in the weight.

        We recall that elements of Drinfeld modular forms ring are not
        necessarily modular forms as they may have mixed weight
        components.

        EXAMPLES::

            sage: A = GF(3)['T']; K = Frac(A)
            sage: M = DrinfeldModularForms(K, 2)
            sage: M.inject_variables()
            Defining g1, g2
            sage: f = g1^5*g2^2  # homogeneous polynomial
            sage: f.is_homogeneous()
            True
            sage: g = g1 + g2  # mixed weight components
            sage: g.is_homogeneous()
            False
        """
        return self._polynomial.is_homogeneous()

    def homogeneous_components(self):
        r"""
        Return the homogeneous components of this graded Drinfeld
        form.

        EXAMPLES::

            sage: A = GF(3)['T']; K = Frac(A)
            sage: M = DrinfeldModularForms(K, 2)
            sage: M.inject_variables()
            Defining g1, g2
            sage: F = g1 + g1^2 + g1*g2^2 + g2^4
            sage: D = F.homogeneous_components(); D
            {2: g1, 4: g1^2, 18: g1*g2^2, 32: g2^4}
            sage: D[32]
            g2^4
        """
        M = self.parent()
        components = self._polynomial.homogeneous_components().items()
        return {k: self.__class__(M, p) for k, p in components}

    def polynomial(self):
        r"""
        Return this graded Drinfeld forms as a multivariate polynomial
        over the generators of the ring.

        OUTPUT: a multivariate polynomial over the base ring

        EXAMPLES::

            sage: A = GF(3)['T']; K = Frac(A)
            sage: M = DrinfeldModularForms(K, 2)
            sage: M.inject_variables()
            Defining g1, g2
            sage: P1 = g1.polynomial();
            sage: P2 = g2.polynomial();
            sage: P2^2 + P1^2 + P1
            g2^2 + g1^2 + g1
            sage: P1.parent()
            Multivariate Polynomial Ring in g1, g2 over Fraction Field of Univariate Polynomial Ring in T over Finite Field of size 3

        The degree of each variables corresponds to the weight of the
        generator::

            sage: P1.degree()
            2
            sage: P2.degree()
            8
        """
        return self._polynomial

    def type(self):
        r"""
        Return the type of this graded Drinfeld form.

        Recall that the *type* is the integer `0 \leq m \leq q-1` such that

        .. MATH::

            f(\gamma(w)) = \mathrm{det}(\gamma)^m j(\gamma, w)^k f(w).

        EXAMPLES::

            sage: A = GF(11)['T']; K = Frac(A)
            sage: M = DrinfeldModularForms(K, 2, has_type=True)
            sage: M.inject_variables()
            Defining g1, h2
            sage: F = g1*h2^9
            sage: F.type()
            9
            sage: (h2^11).type()
            1
            sage: g1.type()
            0

        The type only makes sense when the form is homogeneous::

            sage: F = g1^4 + h2
            sage: F.type()
            Traceback (most recent call last):
            ...
            ValueError: the graded form is not homogeneous

        TESTS::

            sage: A = GF(2)['T']; K = Frac(A)
            sage: M = DrinfeldModularForms(K, 2, has_type=False)
            sage: (M.1).type()
            0
        """
        if not self.is_homogeneous():
            raise ValueError("the graded form is not homogeneous")
        if not self.parent()._has_type:
            return ZZ(0)
        q = self.base_ring().base_ring().cardinality()
        return self.polynomial().degrees()[-1] % (q-1)

    def weight(self):
        r"""
        Return the weight of this graded Drinfeld modular form.

        EXAMPLES::

            sage: A = GF(3)['T']; K = Frac(A)
            sage: M = DrinfeldModularForms(K, 2)
            sage: M.inject_variables()
            Defining g1, g2
            sage: g1.weight()
            2
            sage: g2.weight()
            8
            sage: f = g1^5*g2^2
            sage: f.weight()
            26

        If the form is not homogeneous, then the method returns an
        error::

            sage: f = g1 + g2
            sage: f.weight()
            Traceback (most recent call last):
            ...
            ValueError: the graded form is not homogeneous
        """
        if not self.is_homogeneous():
            raise ValueError("the graded form is not homogeneous")
        return self._polynomial.degree()



# sage.doctest: needs sage.libs.pari
"""
Local and Global Genus Symbols
"""
#############################################################
#                                                           #
#   Wrappers for the Genus/Genus Symbol Code in ../genera/  #
#                                                           #
#############################################################

from sage.arith.misc import is_prime, prime_divisors
from sage.quadratic_forms.genera.genus import Genus, LocalGenusSymbol
from sage.rings.integer_ring import ZZ


def global_genus_symbol(self):
    r"""
    Return the genus of two times a quadratic form over `\ZZ`.

    These are defined by a collection of local genus symbols (a la
    Chapter 15 of Conway-Sloane [CS1999]_), and a signature.

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,2,3,4])
        sage: Q.global_genus_symbol()
        Genus of
        [2 0 0 0]
        [0 4 0 0]
        [0 0 6 0]
        [0 0 0 8]
        Signature:  (4, 0)
        Genus symbol at 2:    [2^-2 4^1 8^1]_6
        Genus symbol at 3:     1^3 3^-1

    ::

        sage: Q = QuadraticForm(ZZ, 4, range(10))
        sage: Q.global_genus_symbol()
        Genus of
        [ 0  1  2  3]
        [ 1  8  5  6]
        [ 2  5 14  8]
        [ 3  6  8 18]
        Signature:  (3, 1)
        Genus symbol at 2:    1^-4
        Genus symbol at 563:     1^3 563^-1
    """
    if self.base_ring() is not ZZ:
        raise TypeError("the quadratic form is not defined over the integers")
    return Genus(self.Hessian_matrix())


def local_genus_symbol(self, p):
    r"""
    Return the Conway-Sloane genus symbol of 2 times a quadratic form
    defined over `\ZZ` at a prime number `p`.

    This is defined (in the class
    :class:`~sage.quadratic_forms.genera.genus.Genus_Symbol_p_adic_ring`)
    to be a list of tuples (one for each Jordan component
    `p^m\cdot A` at `p`, where `A` is a unimodular symmetric matrix with
    coefficients the `p`-adic integers) of the following form:

    - If `p>2`, then return triples of the form [`m`, `n`, `d`] where

      - `m` = valuation of the component

      - `n` = rank of `A`

      - `d` = det(`A`) in {1, `u`} for normalized quadratic non-residue `u`.

    - If `p=2`, then return quintuples of the form [`m`, `n`, `s`, `d`, `o`] where

      - `m` = valuation of the component

      - `n` = rank of `A`

      - `d` = det(`A`) in {1, 3, 5, 7}

      - `s` = 0 (or 1) if `A` is even (or odd)

      - `o` = oddity of `A` (= 0 if `s` = 0) in `\ZZ/8\ZZ` = the trace of the diagonalization of `A`

    .. NOTE::

        The Conway-Sloane convention for describing the prime `p = -1`
        is not supported here, and neither is the convention for
        including the 'prime' Infinity.  See note on p370 of Conway-Sloane
        (3rd ed) [CS1999]_ for a discussion of this convention.

    INPUT:

    - ``p`` -- a prime number > 0

    OUTPUT:

    a Conway-Sloane genus symbol at `p`, which is an
    instance of the class :class:`~sage.quadratic_forms.genera.genus.Genus_Symbol_p_adic_ring`.

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,2,3,4])
        sage: Q.local_genus_symbol(2)
        Genus symbol at 2:    [2^-2 4^1 8^1]_6
        sage: Q.local_genus_symbol(3)
        Genus symbol at 3:     1^3 3^-1
        sage: Q.local_genus_symbol(5)
        Genus symbol at 5:     1^4
    """
    if not is_prime(p):
        raise TypeError("the number " + str(p) + " is not prime")
    if self.base_ring() is not ZZ:
        raise TypeError("the quadratic form is not defined over the integers")
    return LocalGenusSymbol(self.Hessian_matrix(), p)


def CS_genus_symbol_list(self, force_recomputation=False):
    """
    Return the list of Conway-Sloane genus symbols in increasing order of primes dividing 2*det.

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,2,3,4])
        sage: Q.CS_genus_symbol_list()
        [Genus symbol at 2:    [2^-2 4^1 8^1]_6, Genus symbol at 3:     1^3 3^-1]
    """
    # Try to use the cached list
    if not force_recomputation:
        try:
            return self.__CS_genus_symbol_list
        except AttributeError:
            pass

    # Otherwise recompute and cache the list
    list_of_CS_genus_symbols = [self.local_genus_symbol(p)
                                for p in prime_divisors(2 * self.det())]

    self.__CS_genus_symbol_list = list_of_CS_genus_symbols
    return list_of_CS_genus_symbols



# sage.doctest: needs sage.libs.flint sage.libs.pari
"""
Submodules of Hecke modules
"""
# ****************************************************************************
#       Sage: Open Source Mathematical Software
#
#       Copyright (C) 2005 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import sage.arith.misc as arith
from sage.misc.verbose import verbose
from sage.misc.cachefunc import cached_method
from sage.modules.free_module import FreeModule_generic
from sage.structure.richcmp import richcmp_method, richcmp_not_equal

from . import module


@richcmp_method
class HeckeSubmodule(module.HeckeModule_free_module):
    """
    Submodule of a Hecke module.
    """
    def __init__(self, ambient, submodule,
                 dual_free_module=None, check=True) -> None:
        r"""
        Initialise a submodule of an ambient Hecke module.

        INPUT:

        - ``ambient`` -- an ambient Hecke module

        - ``submodule`` -- a free module over the base ring which is a submodule
          of the free module attached to the ambient Hecke module. This should
          be invariant under all Hecke operators.

        - ``dual_free_module`` -- the submodule of the dual of the ambient
          module corresponding to this submodule (or ``None``)

        - ``check`` -- whether or not to explicitly check that the submodule is
          Hecke equivariant

        EXAMPLES::

            sage: CuspForms(1,60) # indirect doctest
            Cuspidal subspace of dimension 5 of Modular Forms space of dimension 6 for Modular Group SL(2,Z) of weight 60 over Rational Field

            sage: M = ModularForms(4,10)
            sage: S = sage.modular.hecke.submodule.HeckeSubmodule(M, M.submodule(M.basis()[:3]).free_module())
            sage: S
            Rank 3 submodule of a Hecke module of level 4

            sage: S == loads(dumps(S))
            True
        """
        from . import ambient_module
        if not isinstance(ambient, ambient_module.AmbientHeckeModule):
            raise TypeError("ambient must be an ambient Hecke module")
        if not isinstance(submodule, FreeModule_generic):
            raise TypeError("submodule must be a free module")
        if not submodule.is_submodule(ambient.free_module()):
            raise ValueError("submodule must be a submodule of the ambient free module")

        if check:
            if not ambient._is_hecke_equivariant_free_module(submodule):
                raise ValueError("The submodule must be invariant under all Hecke operators.")

        self.__ambient = ambient
        self.__submodule = submodule
        module.HeckeModule_free_module.__init__(self, ambient.base_ring(),
                                                ambient.level(),
                                                ambient.weight())
        if dual_free_module is not None:
            if not isinstance(dual_free_module, FreeModule_generic):
                raise TypeError("dual_free_module must be a free module")
            if dual_free_module.rank() != submodule.rank():
                raise ArithmeticError("dual_free_module must have the same rank as submodule")
            self.dual_free_module.set_cache(dual_free_module)

    def _repr_(self):
        r"""
        String representation of ``self``.

        EXAMPLES::

            sage: M = ModularForms(4,10)
            sage: S = sage.modular.hecke.submodule.HeckeSubmodule(M, M.submodule(M.basis()[:3]).free_module())
            sage: S._repr_()
            'Rank 3 submodule of a Hecke module of level 4'
        """
        return "Rank %s submodule of a Hecke module of level %s" % (
            self.rank(), self.level())

    def __add__(self, other):
        r"""
        Sum of ``self`` and ``other`` (as submodules of a common ambient
        module).

        EXAMPLES::

            sage: M = ModularForms(4,10)
            sage: S = sage.modular.hecke.submodule.HeckeSubmodule(M, M.submodule(M.basis()[:3]).free_module())
            sage: E = sage.modular.hecke.submodule.HeckeSubmodule(M, M.submodule(M.basis()[3:]).free_module())
            sage: S + E # indirect doctest
            Modular Forms subspace of dimension 6 of Modular Forms space of dimension 6 for Congruence Subgroup Gamma0(4) of weight 10 over Rational Field
        """
        if not isinstance(other, module.HeckeModule_free_module):
            raise TypeError("other (=%s) must be a Hecke module." % other)
        if self.ambient() != other.ambient():
            raise ArithmeticError("sum only defined for submodules of a common ambient space")
        if other.is_ambient():
            return other
        # Neither is ambient
        M = self.free_module() + other.free_module()
        return self.ambient().submodule(M, check=False)

    def _element_constructor_(self, x, check=True):
        """
        Coerce x into the ambient module and checks that x is in this
        submodule.

        EXAMPLES::

            sage: M = ModularSymbols(37)
            sage: S = M.cuspidal_submodule()
            sage: M([0,oo])
            -(1,0)
            sage: S([0,oo])
            Traceback (most recent call last):
            ...
            TypeError: x does not coerce to an element of this Hecke module
            sage: S([-1/23,0])
            (1,23)
        """
        z = self.ambient_hecke_module()(x).element()
        if check and z not in self.__submodule:
            raise TypeError("x does not coerce to an element of this Hecke module")
        return self.element_class(self, z)

    def __richcmp__(self, other, op):
        """
        Compare ``self`` to ``other``.

        EXAMPLES::

            sage: M = ModularSymbols(12,6)
            sage: S = sage.modular.hecke.submodule.HeckeSubmodule(M, M.cuspidal_submodule().free_module())
            sage: T = sage.modular.hecke.submodule.HeckeSubmodule(M, M.new_submodule().free_module())
            sage: S
            Rank 14 submodule of a Hecke module of level 12
            sage: T
            Rank 0 submodule of a Hecke module of level 12
            sage: S > T
            True
            sage: T < S
            True
            sage: S == S
            True
        """
        if not isinstance(other, module.HeckeModule_free_module):
            return NotImplemented
        lx = self.ambient()
        rx = other.ambient()
        if lx != rx:
            return richcmp_not_equal(lx, rx, op)
        return self.free_module()._echelon_matrix_richcmp(other.free_module(), op)

    ################################
    # Semi-Private functions
    ################################
    def _compute_dual_hecke_matrix(self, n):
        """
        Compute the matrix for the `n`-th Hecke operator acting on
        the dual of ``self``.

        EXAMPLES::

            sage: M = ModularForms(4,10)
            sage: S = sage.modular.hecke.submodule.HeckeSubmodule(M, M.submodule(M.basis()[:3]).free_module())
            sage: S._compute_dual_hecke_matrix(3)
            [    0     0     1]
            [    0  -156     0]
            [35568     0    72]
            sage: CuspForms(4,10).dual_hecke_matrix(3)
            [    0     0     1]
            [    0  -156     0]
            [35568     0    72]
        """
        A = self.ambient_hecke_module().dual_hecke_matrix(n)
        check = arith.gcd(self.level(), n) != 1
        return A.restrict(self.dual_free_module(), check=check)

    def _compute_hecke_matrix(self, n):
        r"""
        Compute the matrix of the `n`-th Hecke operator acting on this space, by
        calling the corresponding function for the ambient space and
        restricting. If `n` is not coprime to the level, we check that the
        restriction is well-defined.

        EXAMPLES::

            sage: R.<q> = QQ[[]]
            sage: M = ModularForms(2, 12)
            sage: f = M(q^2 - 24*q^4 + O(q^6))
            sage: A = M.submodule(M.free_module().span([f.element()]),check=False)
            sage: sage.modular.hecke.submodule.HeckeSubmodule._compute_hecke_matrix(A, 3)
            [252]
            sage: sage.modular.hecke.submodule.HeckeSubmodule._compute_hecke_matrix(A, 4)
            Traceback (most recent call last):
            ...
            ArithmeticError: subspace is not invariant under matrix
        """
        A = self.ambient_hecke_module().hecke_matrix(n)
        check = arith.gcd(self.level(), n) != 1
        return A.restrict(self.free_module(), check=check)

    def _compute_diamond_matrix(self, d):
        r"""
        EXAMPLES::

            sage: f = ModularSymbols(Gamma1(13),2,sign=1).cuspidal_subspace().decomposition()[0]
            sage: a = f.diamond_bracket_operator(2).matrix() # indirect doctest
            sage: a.charpoly()
            x^2 - x + 1
            sage: a^12
            [1 0]
            [0 1]
        """
        return self.ambient_hecke_module().diamond_bracket_matrix(d).restrict(self.free_module(), check=False)

    def _compute_atkin_lehner_matrix(self, d):
        """
        Compute the Atkin-Lehner matrix corresponding to the
        divisor d of the level of ``self``.

        EXAMPLES::

            sage: M = ModularSymbols(4,10)
            sage: S = sage.modular.hecke.submodule.HeckeSubmodule(M, M.cuspidal_submodule().free_module())
            sage: S
            Rank 6 submodule of a Hecke module of level 4
            sage: S._compute_atkin_lehner_matrix(1)
            [1 0 0 0 0 0]
            [0 1 0 0 0 0]
            [0 0 1 0 0 0]
            [0 0 0 1 0 0]
            [0 0 0 0 1 0]
            [0 0 0 0 0 1]
        """
        A = self.ambient_hecke_module()._compute_atkin_lehner_matrix(d)
        return A.restrict(self.free_module(), check=True)

    def _set_dual_free_module(self, V):
        """
        Set the dual free module of ``self`` to V. Here V must be a vector
        space of the same dimension as ``self``, embedded in a space of
        the same dimension as the ambient space of ``self``.

        EXAMPLES::

            sage: M = ModularSymbols(4,10)
            sage: S = sage.modular.hecke.submodule.HeckeSubmodule(M, M.cuspidal_submodule().free_module())
            sage: S._set_dual_free_module(M.cuspidal_submodule().dual_free_module())
            sage: S._set_dual_free_module(S)
        """
        if V.degree() != self.ambient_hecke_module().rank():
            raise ArithmeticError("The degree of V must equal the rank of the ambient space.")
        if V.rank() != self.rank():
            raise ArithmeticError("The rank of V must equal the rank of self.")
        self.dual_free_module.set_cache(V)

    ################################
    # Public functions
    ################################

    def ambient_hecke_module(self):
        r"""
        Return the ambient Hecke module of which this is a submodule.

        EXAMPLES::

            sage: CuspForms(2, 12).ambient_hecke_module()
            Modular Forms space of dimension 4 for Congruence Subgroup Gamma0(2) of weight 12 over Rational Field
        """
        return self.__ambient

    def ambient(self):
        r"""
        Synonym for ambient_hecke_module.

        EXAMPLES::

            sage: CuspForms(2, 12).ambient()
            Modular Forms space of dimension 4 for Congruence Subgroup Gamma0(2) of weight 12 over Rational Field
        """
        return self.__ambient

    @cached_method
    def complement(self, bound=None):
        """
        Return the largest Hecke-stable complement of this space.

        EXAMPLES::

            sage: M = ModularSymbols(15, 6).cuspidal_subspace()
            sage: M.complement()
            Modular Symbols subspace of dimension 4 of Modular Symbols space of dimension 20 for Gamma_0(15) of weight 6 with sign 0 over Rational Field
            sage: E = EllipticCurve("128a")
            sage: ME = E.modular_symbol_space()
            sage: ME.complement()
            Modular Symbols subspace of dimension 17 of Modular Symbols space of dimension 18 for Gamma_0(128) of weight 2 with sign 1 over Rational Field
        """

        if self.dual_free_module.is_in_cache():
            D = self.dual_free_module()
            V = D.basis_matrix().right_kernel()
            return self.submodule(V, check=False)

        if self.is_ambient():
            return self.ambient_hecke_module().zero_submodule()

        if self.is_zero():
            return self.ambient_hecke_module()

        if self.is_full_hecke_module():
            anemic = False
        else:
            anemic = True

        # TODO: optimize in some cases by computing image of
        # complementary factor instead of kernel...?
        verbose("computing")
        N = self.level()
        A = self.ambient_hecke_module()
        V = A.free_module()
        p = 2
        if bound is None:
            bound = A.hecke_bound()
        while True:
            if anemic:
                while N % p == 0:
                    p = arith.next_prime(p)
            verbose("using T_%s" % p)
            f = self.hecke_polynomial(p)
            T = A.hecke_matrix(p)
            g = T.charpoly('x')
            V = T.kernel_on(V, poly=g // f, check=False)
            if V.rank() + self.rank() <= A.rank():
                break
            p = arith.next_prime(p)
            if p > bound:  # to avoid computing Hecke bound unless necessary
                break

        if V.rank() + self.rank() == A.rank():
            return A.submodule(V, check=False)

        # first attempt to compute the complement failed, we now try
        # the following naive approach: decompose the ambient space,
        # decompose self, and sum the pieces of ambient that are not
        # subspaces of self
        verbose("falling back on naive algorithm")
        D = A.decomposition()
        C = A.zero_submodule()
        for X in D:
            if self.intersection(X).dimension() == 0:
                C = C + X
        if C.rank() + self.rank() == A.rank():
            return C

        # failed miserably
        raise RuntimeError("Computation of complementary space failed (cut down to rank %s, but should have cut down to rank %s)." % (
            V.rank(), A.rank() - self.rank()))

    def degeneracy_map(self, level, t=1):
        """
        The `t`-th degeneracy map from ``self`` to the space of ambient modular
        symbols of the given level. The level of ``self`` must be a divisor or
        multiple of level, and `t` must be a divisor of the quotient.

        INPUT:

        - ``level`` -- positive integer; the level of the codomain of the
          map

        - ``t`` -- integer; the parameter of the degeneracy map,
          i.e., the map is related to `f(q)` - `f(q^t)`

        OUTPUT:

        A linear function from ``self`` to the space of modular symbols of given
        level with the same weight, character, sign, etc., as this space.

        EXAMPLES::

            sage: D = ModularSymbols(10,4).cuspidal_submodule().decomposition(); D
            [Modular Symbols subspace of dimension 2 of Modular Symbols space of dimension 10 for Gamma_0(10) of weight 4 with sign 0 over Rational Field,
             Modular Symbols subspace of dimension 4 of Modular Symbols space of dimension 10 for Gamma_0(10) of weight 4 with sign 0 over Rational Field]
            sage: d = D[1].degeneracy_map(5); d
            Hecke module morphism defined by the matrix
            [   0    0   -1    1]
            [   0  1/2  3/2   -2]
            [   0   -1    1    0]
            [   0 -3/4 -1/4    1]
            Domain: Modular Symbols subspace of dimension 4 of Modular Symbols space ...
            Codomain: Modular Symbols space of dimension 4 for Gamma_0(5) of weight ...

        ::

            sage: d.rank()
            2
            sage: d.kernel()
            Modular Symbols subspace of dimension 2 of Modular Symbols space of dimension 10 for Gamma_0(10) of weight 4 with sign 0 over Rational Field
            sage: d.image()
            Modular Symbols subspace of dimension 2 of Modular Symbols space of dimension 4 for Gamma_0(5) of weight 4 with sign 0 over Rational Field
        """
        d = self.ambient_hecke_module().degeneracy_map(level, t)
        return d.restrict_domain(self)

    @cached_method
    def dual_free_module(self, bound=None, anemic=True, use_star=True):
        r"""
        Compute embedded dual free module if possible.

        In general this will not be possible, e.g., if this space is
        not Hecke equivariant, possibly if it is not cuspidal, or if
        the characteristic is not 0. In all these cases we raise a
        :exc:`RuntimeError` exception.

        If ``use_star`` is ``True`` (which is the default), we also use the +/-
        eigenspaces for the star operator to find the dual free module of ``self``.
        If ``self`` does not have a star involution, ``use_star`` will automatically be
        set to ``False``.

        EXAMPLES::

            sage: M = ModularSymbols(11, 2)
            sage: M.dual_free_module()
            Vector space of dimension 3 over Rational Field
            sage: Mpc = M.plus_submodule().cuspidal_submodule()
            sage: Mcp = M.cuspidal_submodule().plus_submodule()
            sage: Mcp.dual_free_module() == Mpc.dual_free_module()
            True
            sage: Mpc.dual_free_module()
            Vector space of degree 3 and dimension 1 over Rational Field
            Basis matrix:
            [  1 5/2   5]

            sage: M = ModularSymbols(35,2).cuspidal_submodule()
            sage: M.dual_free_module(use_star=False)
            Vector space of degree 9 and dimension 6 over Rational Field
            Basis matrix:
            [   1    0    0    0   -1    0    0    4   -2]
            [   0    1    0    0    0    0    0 -1/2  1/2]
            [   0    0    1    0    0    0    0 -1/2  1/2]
            [   0    0    0    1   -1    0    0    1    0]
            [   0    0    0    0    0    1    0   -2    1]
            [   0    0    0    0    0    0    1   -2    1]

            sage: M = ModularSymbols(40,2)
            sage: Mmc = M.minus_submodule().cuspidal_submodule()
            sage: Mcm = M.cuspidal_submodule().minus_submodule()
            sage: Mcm.dual_free_module() == Mmc.dual_free_module()
            True
            sage: Mcm.dual_free_module()
            Vector space of degree 13 and dimension 3 over Rational Field
            Basis matrix:
            [ 0  1  0  0  0  0  1  0 -1 -1  1 -1  0]
            [ 0  0  1  0 -1  0 -1  0  1  0  0  0  0]
            [ 0  0  0  0  0  1  1  0 -1  0  0  0  0]

            sage: M = ModularSymbols(43).cuspidal_submodule()
            sage: S = M[0].plus_submodule() + M[1].minus_submodule()
            sage: S.dual_free_module(use_star=False)
            Traceback (most recent call last):
            ...
            RuntimeError: Computation of complementary space failed (cut down to rank 7, but should have cut down to rank 4).
            sage: S.dual_free_module().dimension() == S.dimension()
            True

        We test that :issue:`5080` is fixed::

            sage: EllipticCurve('128a').congruence_number()
            32
        """

        # if we know the complement we can read off the dual module
        if self.complement.is_in_cache():
            verbose('This module knows its complement already -- cheating in dual_free_module')
            C = self.complement()
            return C.basis_matrix().right_kernel()

        verbose("computing dual")

        A = self.ambient_hecke_module()

        if self.dimension() == 0:
            return A.zero_submodule()

        if A.dimension() == self.dimension():
            return A.free_module()

        # ALGORITHM: Compute the char poly of each Hecke operator on
        # the submodule, then use it to cut out a submodule of the
        # dual.  If the dimension cuts down to the dimension of self
        # terminate with success.  If it stays larger beyond the Sturm
        # bound, raise a RuntimeError exception.

        # In the case that the sign of self is not 1, we need to use
        # the star involution as well as the Hecke operators in order
        # to find the dual of self.
        #
        # Note that one needs to comment out the line caching the
        # result of this computation below in order to get meaningful
        # timings.

        # If the star involution doesn't make sense for self, then we
        # can't use it.
        if not hasattr(self, 'star_eigenvalues'):
            use_star = False

        if use_star:
            # If the star involution has both + and - eigenspaces on self,
            # then we compute the dual on each eigenspace, then put them
            # together.
            if len(self.star_eigenvalues()) == 2:
                V = self.plus_submodule(compute_dual=False).dual_free_module() + \
                    self.minus_submodule(compute_dual=False).dual_free_module()
                return V

            # At this point, we know that self is an eigenspace for star.
            V = A.sign_submodule(self.sign()).dual_free_module()
        else:
            V = A.free_module()

        N = self.level()
        p = 2
        if bound is None:
            bound = A.hecke_bound()
        while True:
            if anemic:
                while N % p == 0:
                    p = arith.next_prime(p)
            verbose("using T_%s" % p)
            f = self.hecke_polynomial(p)
            T = A.dual_hecke_matrix(p)
            V = T.kernel_on(V, poly=f, check=False)
            if V.dimension() <= self.dimension():
                break
            p = arith.next_prime(p)
            if p > bound:
                break

        if V.rank() == self.rank():
            return V

        # Failed to reduce V to the appropriate dimension
        W = self.complement()
        V2 = W.basis_matrix().right_kernel()
        if V2.rank() == self.rank():
            return V2

        raise RuntimeError("Computation of embedded dual vector space failed "
                           "(cut down to rank %s, but should have cut down to rank %s)." % (V.rank(), self.rank()))

    def free_module(self):
        """
        Return the free module corresponding to ``self``.

        EXAMPLES::

            sage: M = ModularSymbols(33,2).cuspidal_subspace() ; M
            Modular Symbols subspace of dimension 6 of Modular Symbols space of dimension 9 for Gamma_0(33) of weight 2 with sign 0 over Rational Field
            sage: M.free_module()
            Vector space of degree 9 and dimension 6 over Rational Field
            Basis matrix:
            [ 0  1  0  0  0  0  0 -1  1]
            [ 0  0  1  0  0  0  0 -1  1]
            [ 0  0  0  1  0  0  0 -1  1]
            [ 0  0  0  0  1  0  0 -1  1]
            [ 0  0  0  0  0  1  0 -1  1]
            [ 0  0  0  0  0  0  1 -1  0]
        """
        return self.__submodule

    def module(self):
        r"""
        Alias for \code{self.free_module()}.

        EXAMPLES::

            sage: M = ModularSymbols(17,4).cuspidal_subspace()
            sage: M.free_module() is M.module()
            True
        """
        return self.free_module()

    def intersection(self, other):
        """
        Return the intersection of ``self`` and ``other``, which must both lie in
        a common ambient space of modular symbols.

        EXAMPLES::

            sage: M = ModularSymbols(43, sign=1)
            sage: A = M[0] + M[1]
            sage: B = M[1] + M[2]
            sage: A.dimension(), B.dimension()
            (2, 3)
            sage: C = A.intersection(B); C.dimension()
            1

        TESTS::

            sage: M = ModularSymbols(1,80)
            sage: M.plus_submodule().cuspidal_submodule().sign() # indirect doctest
            1
        """
        if self.ambient_hecke_module() != other.ambient_hecke_module():
            raise ArithmeticError("intersection only defined for subspaces of"
                                  " a common ambient modular symbols space")
        if other.is_ambient():
            return self
        if self.is_ambient():
            return other

        # Neither is ambient
        V = self.free_module().intersection(other.free_module())
        M = self.ambient_hecke_module().submodule(V, check=False)

        # if sign is nonzero, the intersection will be, too
        # this only makes sense for modular symbols spaces (and hence shouldn't really be in this file)
        try:
            if self.sign():
                M._set_sign(self.sign())
            elif other.sign():
                M._set_sign(other.sign())
        except AttributeError:
            pass

        return M

    def is_ambient(self) -> bool:
        r"""
        Return ``True`` if ``self`` is an ambient space of modular symbols.

        EXAMPLES::

            sage: M = ModularSymbols(17,4)
            sage: M.cuspidal_subspace().is_ambient()
            False
            sage: A = M.ambient_hecke_module()
            sage: S = A.submodule(A.basis())
            sage: sage.modular.hecke.submodule.HeckeSubmodule.is_ambient(S)
            True
        """
        return self.free_module() == self.ambient_hecke_module().free_module()

    def is_new(self, p=None) -> bool:
        """
        Return ``True`` if this Hecke module is `p`-new. If `p` is None,
        returns ``True`` if it is new.

        EXAMPLES::

            sage: M = ModularSymbols(1,16)
            sage: S = sage.modular.hecke.submodule.HeckeSubmodule(M, M.cuspidal_submodule().free_module())
            sage: S.is_new()
            True
        """
        try:
            return self.__is_new[p]
        except AttributeError:
            self.__is_new = {}
        except KeyError:
            pass
        N = self.ambient_hecke_module().new_submodule(p)
        self.__is_new[p] = self.is_submodule(N)
        return self.__is_new[p]

    def is_old(self, p=None) -> bool:
        """
        Return ``True`` if this Hecke module is `p`-old. If `p` is ``None``,
        returns ``True`` if it is old.

        EXAMPLES::

            sage: M = ModularSymbols(50,2)
            sage: S = sage.modular.hecke.submodule.HeckeSubmodule(M, M.old_submodule().free_module())
            sage: S.is_old()
            True
            sage: S = sage.modular.hecke.submodule.HeckeSubmodule(M, M.new_submodule().free_module())
            sage: S.is_old()
            False
        """
        try:
            return self.__is_old[p]
        except AttributeError:
            self.__is_old = {}
        except KeyError:
            pass
        O = self.ambient_hecke_module().old_submodule(p)
        self.__is_old[p] = self.is_submodule(O)
        return self.__is_old[p]

    def is_submodule(self, V) -> bool:
        """
        Return ``True`` if and only if ``self`` is a submodule of V.

        EXAMPLES::

            sage: M = ModularSymbols(30,4)
            sage: S = sage.modular.hecke.submodule.HeckeSubmodule(M, M.cuspidal_submodule().free_module())
            sage: S.is_submodule(M)
            True
            sage: SS = sage.modular.hecke.submodule.HeckeSubmodule(M, M.old_submodule().free_module())
            sage: S.is_submodule(SS)
            False
        """
        if not isinstance(V, module.HeckeModule_free_module):
            return False
        return self.ambient_hecke_module() == V.ambient_hecke_module() and \
            self.free_module().is_subspace(V.free_module())

    def linear_combination_of_basis(self, v):
        """
        Return the linear combination of the basis of ``self`` given
        by the entries of `v`.

        The result can be of different types, and is printed
        accordingly, depending on the type of submodule.

        EXAMPLES::

            sage: M = ModularForms(Gamma0(2),12)

            sage: S = sage.modular.hecke.submodule.HeckeSubmodule(M, M.cuspidal_submodule().free_module())
            sage: S.basis()
            ((1, 0, 0, 0), (0, 1, 0, 0))
            sage: S.linear_combination_of_basis([3, 10])
            (3, 10, 0, 0)

            sage: S = M.cuspidal_submodule()
            sage: S.basis()
            [q + 252*q^3 - 2048*q^4 + 4830*q^5 + O(q^6), q^2 - 24*q^4 + O(q^6)]
            sage: S.linear_combination_of_basis([3, 10])
            3*q + 10*q^2 + 756*q^3 - 6384*q^4 + 14490*q^5 + O(q^6)
        """
        x = self.free_module().linear_combination_of_basis(v)
        return self(x)

    def new_submodule(self, p=None):
        """
        Return the new or `p`-new submodule of this space of modular
        symbols.

        EXAMPLES::

            sage: M = ModularSymbols(20,4)
            sage: M.new_submodule()
            Modular Symbols subspace of dimension 2 of Modular Symbols space of dimension 18 for Gamma_0(20) of weight 4 with sign 0 over Rational Field
            sage: S = sage.modular.hecke.submodule.HeckeSubmodule(M, M.cuspidal_submodule().free_module())
            sage: S
            Rank 12 submodule of a Hecke module of level 20
            sage: S.new_submodule()
            Modular Symbols subspace of dimension 2 of Modular Symbols space of dimension 18 for Gamma_0(20) of weight 4 with sign 0 over Rational Field
        """
        try:
            if self.__is_new[p]:
                return self
        except AttributeError:
            self.__is_new = {}
        except KeyError:
            pass

        if self.rank() == 0:
            self.__is_new[p] = True
            return self
        try:
            return self.__new_submodule[p]
        except AttributeError:
            self.__new_submodule = {}
        except KeyError:
            pass

        S = self.ambient_hecke_module().new_submodule(p)
        ns = S.intersection(self)
        if ns.rank() == self.rank():
            self.__is_new[p] = True
        ns.__is_new = {p: True}
        self.__new_submodule[p] = ns
        return ns

    def nonembedded_free_module(self):
        """
        Return the free module corresponding to ``self`` as an abstract
        free module, i.e. not as an embedded vector space.

        EXAMPLES::

            sage: M = ModularSymbols(12,6)
            sage: S = sage.modular.hecke.submodule.HeckeSubmodule(M, M.cuspidal_submodule().free_module())
            sage: S
            Rank 14 submodule of a Hecke module of level 12
            sage: S.nonembedded_free_module()
            Vector space of dimension 14 over Rational Field
        """
        return self.free_module().nonembedded_free_module()

    def old_submodule(self, p=None):
        r"""
        Return the old or `p`-old submodule of this space of modular
        symbols.

        EXAMPLES: We compute the old and new submodules of
        `\mathbf{S}_2(\Gamma_0(33))`.

        ::

            sage: M = ModularSymbols(33); S = M.cuspidal_submodule(); S
            Modular Symbols subspace of dimension 6 of Modular Symbols space of dimension 9 for Gamma_0(33) of weight 2 with sign 0 over Rational Field
            sage: S.old_submodule()
            Modular Symbols subspace of dimension 4 of Modular Symbols space of dimension 9 for Gamma_0(33) of weight 2 with sign 0 over Rational Field
            sage: S.new_submodule()
            Modular Symbols subspace of dimension 2 of Modular Symbols space of dimension 9 for Gamma_0(33) of weight 2 with sign 0 over Rational Field
        """
        try:
            if self.__is_old[p]:
                return self
        except AttributeError:
            self.__is_old = {}
        except KeyError:
            pass

        if self.rank() == 0:
            self.__is_old[p] = True
            return self
        try:
            return self.__old_submodule[p]
        except AttributeError:
            self.__old_submodule = {}
        except KeyError:
            pass

        S = self.ambient_hecke_module().old_submodule(p)
        os = S.intersection(self)
        if os.rank() == self.rank():
            self.__is_old[p] = True
        os.__is_old = {p: True}
        self.__old_submodule[p] = os
        return os

    def rank(self):
        r"""
        Return the rank of ``self`` as a free module over the base ring.

        EXAMPLES::

            sage: ModularSymbols(6, 4).cuspidal_subspace().rank()
            2
            sage: ModularSymbols(6, 4).cuspidal_subspace().dimension()
            2
        """
        return self.__submodule.rank()

    def submodule(self, M, Mdual=None, check=True):
        """
        Construct a submodule of ``self`` from the free module M, which
        must be a subspace of ``self``.

        EXAMPLES::

            sage: M = ModularSymbols(18,4)
            sage: S = sage.modular.hecke.submodule.HeckeSubmodule(M, M.cuspidal_submodule().free_module())
            sage: S[0]
            Modular Symbols subspace of dimension 2 of Modular Symbols space of dimension 18 for Gamma_0(18) of weight 4 with sign 0 over Rational Field
            sage: S.submodule(S[0].free_module())
            Modular Symbols subspace of dimension 2 of Modular Symbols space of dimension 18 for Gamma_0(18) of weight 4 with sign 0 over Rational Field
        """
        if not isinstance(M, FreeModule_generic):
            V = self.ambient_module().free_module()
            if isinstance(M, (list, tuple)):
                M = V.span([V(x.element()) for x in M])
            else:
                M = V.span(M)

        if check:
            if not M.is_submodule(self.free_module()):
                raise TypeError("M (=%s) must be a submodule of the free module (=%s) associated to this module." % (M, self.free_module()))

        return self.ambient().submodule(M, Mdual, check=check)

    def submodule_from_nonembedded_module(self, V, Vdual=None, check=True):
        """
        Construct a submodule of ``self`` from V. Here V should be a
        subspace of a vector space whose dimension is the same as that
        of ``self``.

        INPUT:

        - ``V`` -- submodule of ambient free module of the same
          rank as the rank of ``self``

        - ``check`` -- whether to check that V is Hecke
          equivariant

        OUTPUT: Hecke submodule of self

        EXAMPLES::

            sage: M = ModularSymbols(37,2)
            sage: S = sage.modular.hecke.submodule.HeckeSubmodule(M, M.cuspidal_submodule().free_module())
            sage: V = (QQ**4).subspace([[1,-1,0,1/2],[0,0,1,-1/2]])
            sage: S.submodule_from_nonembedded_module(V)
            Modular Symbols subspace of dimension 2 of Modular Symbols space of dimension 5 for Gamma_0(37) of weight 2 with sign 0 over Rational Field
        """
        E = self.free_module()
        M_V = V.matrix()
        M_E = E.matrix()
        # We encode the operation of taking the linear combinations of
        # the basis of E given by the basis of V as a single matrix
        # multiplication, since matrix multiplication is (presumed to be)
        # so fast, and their are asymptotically fast algorithms.
        A = M_V * M_E
        V = A.row_space()
        if Vdual is not None:
            E = self.dual_free_module()
            M_Vdual = Vdual.matrix()
            M_E = E.matrix()
            A = M_Vdual * M_E
            Vdual = A.row_space()
        return self.ambient_hecke_module().submodule(V, Vdual, check=check)

    def hecke_bound(self):
        r"""
        Compute the Hecke bound for ``self``.

        This is a number `n` such that the `T_m` for `m \leq n`
        generate the Hecke algebra.

        EXAMPLES::

            sage: M = ModularSymbols(24,8)
            sage: M.hecke_bound()
            53
            sage: M.cuspidal_submodule().hecke_bound()
            32
            sage: M.eisenstein_submodule().hecke_bound()
            53
        """
        if self.is_cuspidal():
            return self.sturm_bound()
        return self.ambient_hecke_module().hecke_bound()



from sage.rings.tate_algebra_element cimport TateAlgebraTerm
from sage.rings.tate_algebra_element cimport TateAlgebraElement

cdef Jpair(p1, p2)
cdef TateAlgebraElement regular_reduce(sgb, TateAlgebraTerm s, TateAlgebraElement v, stopval)
cdef TateAlgebraElement reduce(gb, TateAlgebraElement v, stopval)



from sage.libs.ntl.types cimport ZZX_c
from sage.libs.mpfr.types cimport mpfr_t
from sage.libs.mpfi.types cimport mpfi_t

cdef ZZX_evaluation_mpfr(mpfr_t res, ZZX_c poly, const mpfr_t a)
cdef ZZX_evaluation_mpfi(mpfi_t res, ZZX_c poly, const mpfi_t a)



from sage.misc.lazy_import import lazy_import

from sage.modular.quatalg.all import *

from sage.modular.modsym.all import *

from sage.modular.modform.all import *

from sage.modular.ssmod.all import *

from sage.modular.abvar.all import *

from sage.modular.dirichlet import (DirichletGroup,
                                    kronecker_character, kronecker_character_upside_down,
                                    trivial_character)

from sage.modular.arithgroup.all import (Gamma0, Gamma1, GammaH, Gamma, SL2Z,
                                         ArithmeticSubgroup_Permutation,
                                         CongruenceSubgroup, FareySymbol)

from sage.modular.cusps import Cusp, Cusps

from sage.modular.etaproducts import (EtaGroup, EtaProduct, EtaGroupElement,
                                      AllCusps, CuspFamily)

lazy_import('sage.modular.multiple_zeta', ['Multizeta', 'Multizetas'])

from sage.modular.overconvergent.all import *

from sage.modular.local_comp.all import *

from sage.modular.cusps_nf import NFCusp, NFCusps, Gamma0_NFCusps

from sage.modular.btquotients.all import *

from sage.modular.pollack_stevens.all import *

from sage.modular.quasimodform.all import *

from sage.modular.drinfeld_modform.all import *

del lazy_import



"""
Kummer surfaces over a general ring

AUTHORS:

- David Kohel (2006): initial version
- Sabrina Kunzweiler, Gareth Ma, Giacomo Pope (2026): add map from the Jacobian
"""

# *****************************************************************************
#  Copyright (C) 2006 David Kohel <kohel@maths.usyd.edu>
#                2026 Sabrina Kunzweiler <sabrina.kunzweiler@math.u-bordeaux.fr>
#                2026 Gareth Ma <grhkm21@gmail.com>
#                2026 Giacomo Pope <giacomopope@gmail.com>
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
# *****************************************************************************

from sage.schemes.hyperelliptic_curves.jacobian_morphism import MumfordDivisorClassField
from sage.schemes.projective.projective_space import ProjectiveSpace
from sage.schemes.projective.projective_subscheme import (
    AlgebraicScheme_subscheme_projective,
)


class KummerSurface(AlgebraicScheme_subscheme_projective):
    r"""
    Kummer surface of the Jacobian of a genus-2 curve.

    EXAMPLES::

        sage: R.<x> = GF(13)[]
        sage: C = HyperellipticCurve(x^6+2*x+1)
        sage: J = C.jacobian()
        sage: K = J.kummer_surface()

    Points can be constructed by providing explicit coordinates,
    or as the images of elements of the Jacobian::

        sage: P = K([3,6,0,1]); P
        (3 : 6 : 0 : 1)
        sage: Q = J([x^2+10*x+2,11])
        sage: K(Q)
        (9 : 1 : 5 : 1)
    """

    def __init__(self, J):
        r"""
        Constructor for a Kummer surface of the Jacobian of a genus-2 curve.

        The equation for the Kummer surface is based on the code provided in
        https://people.maths.ox.ac.uk/flynn/genus2/kummer/ with the modifications
        outlined in [Mue2010]_.

        EXAMPLES::

            sage: R.<x> = QQ[]
            sage: f = x^5 + x + 1
            sage: C = HyperellipticCurve(f)
            sage: J = Jacobian(C)
            sage: K = KummerSurface(J); K
            Kummer Surface of Jacobian of Hyperelliptic Curve over Rational Field defined by y^2 = x^5 + x + 1.
            The defining equation is X0^4 - 4*X0*X1^3 + 4*X0^2*X1*X2 - 4*X0*X1^2*X2 + 2*X0^2*X2^2 + X2^4 - 4*X0^3*X3 - 2*X0^2*X1*X3 - 2*X1*X2^2*X3 + X1^2*X3^2 - 4*X0*X2*X3^2
        """
        R = J.base_ring()
        self._base_ring = R
        self._curve = J.curve()
        self._jacobian = J

        PP = ProjectiveSpace(3, R, ["X0", "X1", "X2", "X3"])
        X0, X1, X2, X3 = PP.gens()
        f, h = self._curve.hyperelliptic_polynomials()
        [h0, h1, h2, h3] = [h[i] for i in range(4)]
        [f0, f1, f2, f3, f4, f5, f6] = [f[i] for i in range(7)]

        K2 = X1**2 - 4 * X0 * X2

        K1 = (
            (4 * f0 + h0**2) * X0**3
            + (2 * f1 + h0 * h1) * X0**2 * X1
            + (h0 * h2) * X0 * X1**2
            + (h0 * h3) * X1**3
            + (4 * f2 - 2 * h0 * h2 + h1**2) * X0**2 * X2
            + (2 * f3 - 3 * h0 * h3 + h1 * h2) * X0 * X1 * X2
            + (h1 * h3) * X1**2 * X2
            + (4 * f4 - 2 * h1 * h3 + h2**2) * X0 * X2**2
            + (2 * f5 + h2 * h3) * X1 * X2**2
            + (4 * f6 + h3**2) * X2**3
        )

        K0 = (
            (-4 * f0 * f2 - f0 * h1**2 + f1**2 + f1 * h0 * h1 - f2 * h0**2) * X0**4
            + (-4 * f0 * f3 - 2 * f0 * h1 * h2 + f1 * h0 * h2 - f3 * h0**2) * X0**3 * X1
            + (-4 * f0 * f4 - 2 * f0 * h1 * h3 - f0 * h2**2 + f1 * h0 * h3 - f4 * h0**2)
            * X0**2
            * X1**2
            + (-4 * f0 * f5 - 2 * f0 * h2 * h3 - f5 * h0**2) * X0 * X1**3
            + (-4 * f0 * f6 - f0 * h3**2 - f6 * h0**2) * X1**4
            + (
                2 * f0 * h1 * h3
                - 2 * f1 * f3
                - f1 * h0 * h3
                - f1 * h1 * h2
                + 2 * f2 * h0 * h2
                - f3 * h0 * h1
            )
            * X0**3
            * X2
            + (
                4 * f0 * f5
                + 2 * f0 * h2 * h3
                - 4 * f1 * f4
                - f1 * h1 * h3
                - f1 * h2**2
                + 2 * f2 * h0 * h3
                + f3 * h0 * h2
                - 2 * f4 * h0 * h1
                + f5 * h0**2
            )
            * X0**2
            * X1
            * X2
            + (
                8 * f0 * f6
                + 2 * f0 * h3**2
                - 4 * f1 * f5
                - 2 * f1 * h2 * h3
                + f3 * h0 * h3
                - 2 * f5 * h0 * h1
                + 2 * f6 * h0**2
            )
            * X0
            * X1**2
            * X2
            + (-4 * f1 * f6 - f1 * h3**2 - 2 * f6 * h0 * h1) * X1**3 * X2
            + (
                -4 * f0 * f6
                - f0 * h3**2
                + 2 * f1 * f5
                + f1 * h2 * h3
                - 4 * f2 * f4
                - f2 * h2**2
                + f3**2
                + f3 * h0 * h3
                + f3 * h1 * h2
                - f4 * h1**2
                + f5 * h0 * h1
                - f6 * h0**2
            )
            * X0**2
            * X2**2
            + (
                4 * f1 * f6
                + f1 * h3**2
                - 4 * f2 * f5
                - 2 * f2 * h2 * h3
                + f3 * h1 * h3
                + 2 * f4 * h0 * h3
                - f5 * h0 * h2
                - f5 * h1**2
                + 2 * f6 * h0 * h1
            )
            * X0
            * X1
            * X2**2
            + (-4 * f2 * f6 - f2 * h3**2 + f5 * h0 * h3 - 2 * f6 * h0 * h2 - f6 * h1**2)
            * X1**2
            * X2**2
            + (
                -2 * f3 * f5
                - f3 * h2 * h3
                + 2 * f4 * h1 * h3
                - f5 * h0 * h3
                - f5 * h1 * h2
                + 2 * f6 * h0 * h2
            )
            * X0
            * X2**3
            + (-4 * f3 * f6 - f3 * h3**2 + f5 * h1 * h3 - 2 * f6 * h1 * h2) * X1 * X2**3
            + (-4 * f4 * f6 - f4 * h3**2 + f5**2 + f5 * h2 * h3 - f6 * h2**2) * X2**4
        )

        F = K2 * X3**2 - K1 * X3 + K0
        self._defining_equation = F
        AlgebraicScheme_subscheme_projective.__init__(self, PP, F)

        J._kummer_surface = self

    def __repr__(self):
        r"""
        String representation of the Kummer surface.

        EXAMPLES::

            sage: R.<x> = GF(19)[]
            sage: C = HyperellipticCurve(x^5-1)
            sage: J = Jacobian(C)
            sage: K = KummerSurface(J); K
            Kummer Surface of Jacobian of Hyperelliptic Curve over Finite Field of size 19 defined by y^2 = x^5 + 18.
            The defining equation is 4*X0*X1^3 - 4*X0^2*X1*X2 + X2^4 + 4*X0^3*X3 - 2*X1*X2^2*X3 + X1^2*X3^2 - 4*X0*X2*X3^2
        """
        return f"Kummer Surface of {self._jacobian}.\nThe defining equation is {self._defining_equation}"

    def _mumford_to_kummer(self, P):
        r"""
        Given a point P on the Jacobian J of a genus-2 curve,
        return the Kummer coordinates of P.

        TESTS::

            sage: R.<x> = GF(19)[]
            sage: H = HyperellipticCurve(x^6 + 3*x + 5, x^2 + x)
            sage: J = Jacobian(H)
            sage: K = J.kummer_surface()
            sage: P = J([x^2 + 11*x + 13, 18*x + 10])
            sage: K._mumford_to_kummer(P)
            (7, 18, 15, 15)
            sage: Q = J([x, 9])
            sage: K._mumford_to_kummer(Q)
            (0, 1, 0, 18)
            sage: K._mumford_to_kummer(J.zero())
            (0, 0, 0, 1)
            sage: S = J.random_element()
            sage: K(S) == K(-S)
            True
        """

        J = self.jacobian()
        C = J.curve()
        R = J.base_ring()
        f, h = C.hyperelliptic_polynomials()
        [h0, h1, h2, h3] = [h[i] for i in range(4)]
        [f0, f1, f2, f3, f4, f5, f6] = [f[i] for i in range(7)]
        u, v = P.uv()
        [u0, u1, u2] = [u[i] for i in range(3)]
        [v0, v1] = [v[i] for i in range(2)]

        # divisors of the form [P + Q - D_infty]
        if u2 == R.one():
            # auxiliary values
            F0 = (
                2 * u0**3 * f6
                - u0**2 * u1 * f5
                + 2 * u0**2 * f4
                - u0 * u1 * f3
                + 2 * u0 * f2
                - u1 * f1
                + 2 * f0
            )
            y1y2 = u0 * v1**2 - u1 * v0 * v1 + v0**2
            h1h2 = (
                h3**2 * u0**3
                + (-h2 * h3) * u0**2 * u1
                + (h2**2 - 2 * h1 * h3) * u0**2
                + h1 * h3 * u0 * u1**2
                + (-h1 * h2 + 3 * h0 * h3) * u0 * u1
                + (h1**2 - 2 * h0 * h2) * u0
                + (-h0 * h3) * u1**3
                + h0 * h2 * u1**2
                + (-h0 * h1) * u1
                + h0**2
            )

            if u1**2 - 4 * u0 == 0:
                # divisor of the form [2*P - D_infty] with P = [-u1/2,v0]
                # TODO: Is this formula correct?
                denom = 4 * (
                    (4 * f6 + h3**2) * u1**6
                    + (-8 * f5 - 4 * h2 * h3) * u1**5
                    + (16 * f4 + 8 * h1 * h3 + 4 * h2**2) * u1**4
                    + (-32 * f3 - 16 * h0 * h3 - 16 * h1 * h2) * u1**3
                    + (64 * f2 + 32 * h0 * h2 + 16 * h1**2) * u1**2
                    + (-128 * f1 - 64 * h0 * h1) * u1
                    + (256 * f0 + 64 * h0**2)
                )
                x0 = denom
                x1 = -u1 * denom
                x2 = u0 * denom
                x3 = (-1) * (
                    (4 * f4 * f6 + f4 * h3**2 - f5**2 - f5 * h2 * h3 + f6 * h2**2)
                    * u1**8
                    + (
                        -16 * f3 * f6
                        - 4 * f3 * h3**2
                        + 4 * f5 * h1 * h3
                        - 8 * f6 * h1 * h2
                    )
                    * u1**7
                    + (
                        64 * f2 * f6
                        + 16 * f2 * h3**2
                        + 8 * f3 * f5
                        + 4 * f3 * h2 * h3
                        - 8 * f4 * h1 * h3
                        - 12 * f5 * h0 * h3
                        + 4 * f5 * h1 * h2
                        + 24 * f6 * h0 * h2
                        + 16 * f6 * h1**2
                    )
                    * u1**6
                    + (
                        -192 * f1 * f6
                        - 48 * f1 * h3**2
                        - 64 * f2 * f5
                        - 32 * f2 * h2 * h3
                        + 16 * f3 * h1 * h3
                        + 32 * f4 * h0 * h3
                        - 16 * f5 * h0 * h2
                        - 16 * f5 * h1**2
                        - 96 * f6 * h0 * h1
                    )
                    * u1**5
                    + (
                        576 * f0 * f6
                        + 144 * f0 * h3**2
                        + 224 * f1 * f5
                        + 112 * f1 * h2 * h3
                        + 64 * f2 * f4
                        + 16 * f2 * h2**2
                        - 16 * f3**2
                        - 80 * f3 * h0 * h3
                        - 16 * f3 * h1 * h2
                        + 16 * f4 * h1**2
                        + 112 * f5 * h0 * h1
                        + 144 * f6 * h0**2
                    )
                    * u1**4
                    + (
                        -768 * f0 * f5
                        - 384 * f0 * h2 * h3
                        - 256 * f1 * f4
                        - 64 * f1 * h1 * h3
                        - 64 * f1 * h2**2
                        + 128 * f2 * h0 * h3
                        + 64 * f3 * h0 * h2
                        - 128 * f4 * h0 * h1
                        - 192 * f5 * h0**2
                    )
                    * u1**3
                    + (
                        1024 * f0 * f4
                        + 384 * f0 * h1 * h3
                        + 256 * f0 * h2**2
                        + 128 * f1 * f3
                        - 192 * f1 * h0 * h3
                        + 64 * f1 * h1 * h2
                        - 128 * f2 * h0 * h2
                        + 64 * f3 * h0 * h1
                        + 256 * f4 * h0**2
                    )
                    * u1**2
                    + (
                        -1024 * f0 * f3
                        - 512 * f0 * h1 * h2
                        + 256 * f1 * h0 * h2
                        - 256 * f3 * h0**2
                    )
                    * u1
                    + (
                        1024 * f0 * f2
                        + 256 * f0 * h1**2
                        - 256 * f1**2
                        - 256 * f1 * h0 * h1
                        + 256 * f2 * h0**2
                    )
                )
            elif y1y2 == R.zero():
                # divisor of the form [(a,0) + (b,0) - D_infty]
                denom = u1**2 - 4 * u0
                x0 = denom
                x1 = -u1 * denom
                x2 = u0 * denom
                x3 = F0 + h1h2 - y1y2
            else:
                # the generic case
                denom = 4 * y1y2 * (u1**2 - 4 * u0)
                x0 = denom
                x1 = -u1 * denom
                x2 = u0 * denom
                term = (
                    (-(f5**2) + 4 * f4 * f6) * u0**4
                    + (-4 * f3 * f6) * u0**3 * u1
                    + 2 * f3 * f5 * u0**3
                    + 4 * f2 * f6 * u0**2 * u1**2
                    + (-4 * f2 * f5 + 4 * f1 * f6) * u0**2 * u1
                    + (-(f3**2) + 4 * f2 * f4 - 2 * f1 * f5 + 4 * f0 * f6) * u0**2
                    + (-4 * f1 * f6) * u0 * u1**3
                    + (4 * f1 * f5 - 8 * f0 * f6) * u0 * u1**2
                    + (-4 * f1 * f4 + 4 * f0 * f5) * u0 * u1
                    + 2 * f1 * f3 * u0
                    + 4 * f0 * f6 * u1**4
                    + (-4 * f0 * f5) * u1**3
                    + 4 * f0 * f4 * u1**2
                    + (-4 * f0 * f3) * u1
                    - f1**2
                    + 4 * f0 * f2
                )
                x3 = 4 * (F0 + h1h2 - y1y2) * y1y2 - (term * (u1**2 - 4 * u0) + F0**2)
        elif u1 == R.one():
            # In this case, the divisor is of the form [P + O0 - D_{\infty}],
            # where P = (-u0,v0) and O0 is a point at infinity.
            if C.is_ramified():
                O0 = C.points_at_infinity()[0]
            if C.is_split():
                [P_plus, P_minus] = C.points_at_infinity()
                if P._n == 1:
                    O0 = P_plus
                else:
                    O0 = P_minus
            # coordinates of the points in the support of the divisor
            X1, Y1, Z1 = -u0, v0, R.one()
            X2, Y2, Z2 = O0

            x0 = R.zero()
            x1 = X2**3 * Z1**3
            x2 = X1 * X2**3 * Z1**2
            x3 = -(
                2 * Y1 * Y2
                + h3 * Y1 * X2**3
                + Y2 * (h3 * X1**3 + h2 * X1**2 * Z1 + h1 * X1 * Z1**2 + h0 * Z1**3)
                - 2 * f6 * X1**3 * X2**3
                - f5 * X1**2 * X2**3 * Z1
            )
        elif C.is_split() and P._n != 1:
            # divisor [P_infty+ - P_infty-]
            # TODO: Is this the correct image?
            x0 = R.zero()
            x1 = R.zero()
            x2 = -(h3**2) - 4 * f6
            x3 = f6 * h2**2 - f5 * h2 * h3 + f4 * h3**2 - f5**2 + 4 * f4 * f6
        else:  # P = 0
            x0, x1, x2, x3 = R.zero(), R.zero(), R.zero(), R.one()
        return (x0, x1, x2, x3)

    def __call__(self, P):
        r"""
        Create a point on the Kummer surface.

        INPUT: ``P`` -- either a point `P` on the Jacobian or
        coordinates `(X0, X1, X2, X3)` defining a point on the Kummer surface

        OUTPUT: A point on the Kummer surface.

        EXAMPLES::

            sage: R.<x> = FiniteField(13)[]
            sage: H = HyperellipticCurve(x^6 + 11*x^4 + 6*x^3 + 10*x^2 + 11*x + 1)
            sage: J = Jacobian(H)
            sage: K = KummerSurface(J)
            sage: D1 = K([1,11,9,1]); D1
            (1 : 11 : 9 : 1)
            sage: P1 = J(x^2 + 2*x + 9, 4*x + 4)
            sage: K(P1)
            (1 : 11 : 9 : 1)
        """
        if P == 0:
            coords = (0, 0, 0, 1)
        # construct point from Mumford coordinates
        elif isinstance(P, MumfordDivisorClassField):
            coords = self._mumford_to_kummer(P)
        else:
            if not len(P) == 4:
                raise ValueError("The input must consist of 4 coordinates")
            coords = P

        return self.point(coords)

    def defining_equation(self):
        r"""
        Return the defining equation of the Kummer surface in P^3.

        EXAMPLES::

            sage: R.<x> = GF(19)[]
            sage: C = HyperellipticCurve(x^5-1)
            sage: J = Jacobian(C)
            sage: K = KummerSurface(J); K.defining_equation()
            4*X0*X1^3 - 4*X0^2*X1*X2 + X2^4 + 4*X0^3*X3 - 2*X1*X2^2*X3 + X1^2*X3^2 - 4*X0*X2*X3^2
            sage: K.ambient_space()
            Projective Space of dimension 3 over Finite Field of size 19
        """
        return self._defining_equation

    def jacobian(self):
        r"""
        Return the Jacobain of the genus-2 curve lying above the Kummer surface.

        EXAMPLES::

            sage: R.<x> = GF(19)[]
            sage: C = HyperellipticCurve(x^5-1)
            sage: J = Jacobian(C)
            sage: K = KummerSurface(J); K.jacobian()
            Jacobian of Hyperelliptic Curve over Finite Field of size 19 defined by y^2 = x^5 + 18
        """
        return self._jacobian



from sage.misc.lazy_import import lazy_import

from sage.groups.pari_group import PariGroup

from sage.groups.matrix_gps.all import *
from sage.groups.abelian_gps.all import *

from sage.groups.perm_gps.all import *

from sage.groups.generic import (discrete_log, discrete_log_rho, discrete_log_lambda,
                      linear_relation, multiple, multiples, order_from_multiple)

lazy_import('sage.groups.class_function', 'ClassFunction')

from sage.groups.additive_abelian.all import *

lazy_import('sage.groups.conjugacy_classes', ['ConjugacyClass', 'ConjugacyClassGAP'])

lazy_import('sage.groups.free_group', 'FreeGroup')
lazy_import('sage.groups.braid', 'BraidGroup')
lazy_import('sage.groups.cubic_braid', 'CubicBraidGroup')
lazy_import('sage.groups.cubic_braid', 'AssionGroupU')
lazy_import('sage.groups.cubic_braid', 'AssionGroupS')

lazy_import('sage.groups.affine_gps.affine_group', 'AffineGroup')
lazy_import('sage.groups.affine_gps.euclidean_group', 'EuclideanGroup')

lazy_import('sage.groups.artin', 'ArtinGroup')
lazy_import('sage.groups.raag', 'RightAngledArtinGroup')

lazy_import('sage.groups', 'groups_catalog', 'groups')

lazy_import('sage.groups.semimonomial_transformations.semimonomial_transformation_group', 'SemimonomialTransformationGroup')

lazy_import('sage.groups.group_exp', 'GroupExp')
lazy_import('sage.groups.group_exp', ['GroupExp_Class', 'GroupExpElement'],
            deprecation=38238)

lazy_import('sage.groups.group_semidirect_product', 'GroupSemidirectProduct')
lazy_import('sage.groups.group_semidirect_product', 'GroupSemidirectProductElement',
            deprecation=38238)



# sage.doctest: needs sage.combinat sage.graphs
"""
Dynkin diagrams

AUTHORS:

- Travis Scrimshaw (2012-04-22): Nicolas M. Thiery moved Cartan matrix creation
  to here and I cached results for speed.

- Travis Scrimshaw (2013-06-11): Changed inputs of Dynkin diagrams to handle
  other Dynkin diagrams and graphs. Implemented remaining Cartan type methods.

- Christian Stump, Travis Scrimshaw (2013-04-11): Added Cartan matrix as
  possible input for Dynkin diagrams.
"""
# ****************************************************************************
#       Copyright (C) 2007 Mike Hansen <mhansen@gmail.com>,
#       Copyright (C) 2013 Travis Scrimshaw <tscrim@ucdavis.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from sage.misc.cachefunc import cached_method
from sage.misc.lazy_import import lazy_import
from sage.structure.element import Matrix
from sage.graphs.digraph import DiGraph
from sage.combinat.root_system.cartan_type import CartanType, CartanType_abstract

lazy_import('sage.combinat.root_system.cartan_matrix', 'CartanMatrix')


def DynkinDiagram(*args, **kwds):
    r"""
    Return the Dynkin diagram corresponding to the input.

    INPUT:

    The input can be one of the following:

    - empty to obtain an empty Dynkin diagram
    - a Cartan type
    - a Cartan matrix
    - a Cartan matrix and an indexing set

    One can also input an indexing set by passing a tuple using the optional
    argument ``index_set``.

    The edge multiplicities are encoded as edge labels. For the corresponding
    Cartan matrices, this uses the convention in Hong and Kang, Kac,
    Fulton and Harris, and crystals. This is the **opposite** convention
    in Bourbaki and Wikipedia's Dynkin diagram (:wikipedia:`Dynkin_diagram`).
    That is for `i \neq j`::

        i <--k-- j <==> a_ij = -k
                   <==> -scalar(coroot[i], root[j]) = k
                   <==> multiple arrows point from the longer root
                        to the shorter one

    For example, in type `C_2`, we have::

        sage: C2 = DynkinDiagram(['C',2]); C2
        O=<=O
        1   2
        C2
        sage: C2.cartan_matrix()
        [ 2 -2]
        [-1  2]

    However Bourbaki would have the Cartan matrix as:

    .. MATH::

        \begin{bmatrix}
        2 & -1 \\
        -2 & 2
        \end{bmatrix}.

    EXAMPLES::

        sage: DynkinDiagram(['A', 4])
        O---O---O---O
        1   2   3   4
        A4

        sage: DynkinDiagram(['A',1],['A',1])
        O
        1
        O
        2
        A1xA1

        sage: R = RootSystem("A2xB2xF4")
        sage: DynkinDiagram(R)
        O---O
        1   2
        O=>=O
        3   4
        O---O=>=O---O
        5   6   7   8
        A2xB2xF4

        sage: R = RootSystem("A2xB2xF4")
        sage: CM = R.cartan_matrix(); CM
        [ 2 -1| 0  0| 0  0  0  0]
        [-1  2| 0  0| 0  0  0  0]
        [-----+-----+-----------]
        [ 0  0| 2 -1| 0  0  0  0]
        [ 0  0|-2  2| 0  0  0  0]
        [-----+-----+-----------]
        [ 0  0| 0  0| 2 -1  0  0]
        [ 0  0| 0  0|-1  2 -1  0]
        [ 0  0| 0  0| 0 -2  2 -1]
        [ 0  0| 0  0| 0  0 -1  2]
        sage: DD = DynkinDiagram(CM); DD
        O---O
        1   2
        O=>=O
        3   4
        O---O=>=O---O
        5   6   7   8
        A2xB2xF4
        sage: DD.cartan_matrix()
        [ 2 -1  0  0  0  0  0  0]
        [-1  2  0  0  0  0  0  0]
        [ 0  0  2 -1  0  0  0  0]
        [ 0  0 -2  2  0  0  0  0]
        [ 0  0  0  0  2 -1  0  0]
        [ 0  0  0  0 -1  2 -1  0]
        [ 0  0  0  0  0 -2  2 -1]
        [ 0  0  0  0  0  0 -1  2]

    We can also create Dynkin diagrams from arbitrary Cartan matrices::

        sage: C = CartanMatrix([[2, -3], [-4, 2]])
        sage: DynkinDiagram(C)
        Dynkin diagram of rank 2
        sage: C.index_set()
        (0, 1)
        sage: CI = CartanMatrix([[2, -3], [-4, 2]], [3, 5])
        sage: DI = DynkinDiagram(CI)
        sage: DI.index_set()
        (3, 5)
        sage: CII = CartanMatrix([[2, -3], [-4, 2]])
        sage: DII = DynkinDiagram(CII, ('y', 'x'))
        sage: DII.index_set()
        ('x', 'y')

    .. SEEALSO::

        :func:`CartanType` for a general discussion on Cartan
        types and in particular node labeling conventions.

    TESTS:

    Check that :issue:`15277` is fixed by not having edges from 0s::

        sage: CM = CartanMatrix([[2,-1,0,0],[-3,2,-2,-2],[0,-1,2,-1],[0,-1,-1,2]])
        sage: CM
        [ 2 -1  0  0]
        [-3  2 -2 -2]
        [ 0 -1  2 -1]
        [ 0 -1 -1  2]
        sage: CM.dynkin_diagram().edges(sort=True)
        [(0, 1, 3),
         (1, 0, 1),
         (1, 2, 1),
         (1, 3, 1),
         (2, 1, 2),
         (2, 3, 1),
         (3, 1, 2),
         (3, 2, 1)]
    """
    if len(args) == 0:
        return DynkinDiagram_class()
    mat = args[0]
    if isinstance(mat, Matrix):
        mat = CartanMatrix(*args)
    if isinstance(mat, CartanMatrix):
        if mat.cartan_type() is not mat:
            try:
                return mat.cartan_type().dynkin_diagram()
            except AttributeError:
                ct = CartanType(*args)
                raise ValueError("Dynkin diagram data not yet hardcoded for type %s" % ct)
        if len(args) > 1:
            index_set = tuple(args[1])
        elif "index_set" in kwds:
            index_set = tuple(kwds["index_set"])
        else:
            index_set = mat.index_set()
        D = DynkinDiagram_class(index_set=index_set)
        for (i, j) in mat.nonzero_positions():
            if i != j:
                D.add_edge(index_set[i], index_set[j], -mat[j, i])
        return D
    ct = CartanType(*args)
    try:
        return ct.dynkin_diagram()
    except AttributeError:
        raise ValueError("Dynkin diagram data not yet hardcoded for type %s" % ct)


class DynkinDiagram_class(DiGraph, CartanType_abstract):
    """
    A Dynkin diagram.

    .. SEEALSO::

        :func:`DynkinDiagram()`

    INPUT:

    - ``t`` -- a Cartan type, Cartan matrix, or ``None``

    EXAMPLES::

        sage: DynkinDiagram(['A', 3])
        O---O---O
        1   2   3
        A3
        sage: C = CartanMatrix([[2, -3], [-4, 2]])
        sage: DynkinDiagram(C)
        Dynkin diagram of rank 2
        sage: C.dynkin_diagram().cartan_matrix() == C
        True

    TESTS:

    Check that the correct type is returned when copied::

        sage: d = DynkinDiagram(['A', 3])
        sage: type(copy(d))
        <class 'sage.combinat.root_system.dynkin_diagram.DynkinDiagram_class'>

    We check that :issue:`14655` is fixed::

        sage: cd = copy(d)
        sage: cd.add_vertex(4)
        sage: d.vertices(sort=True) != cd.vertices(sort=True)
        True

    Implementation note: if a Cartan type is given, then the nodes
    are initialized from the index set of this Cartan type.
    """

    def __init__(self, t=None, index_set=None, odd_isotropic_roots=[],
                 **options):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: d = DynkinDiagram(["A", 3])
            sage: TestSuite(d).run()
        """
        if isinstance(t, DiGraph):
            if isinstance(t, DynkinDiagram_class):
                self._cartan_type = t._cartan_type
                self._odd_isotropic_roots = tuple(odd_isotropic_roots)
            else:
                self._cartan_type = None
                self._odd_isotropic_roots = ()
            DiGraph.__init__(self, data=t, **options)
            return

        DiGraph.__init__(self, **options)
        self._cartan_type = t
        self._odd_isotropic_roots = tuple(odd_isotropic_roots)
        if index_set is not None:
            self.add_vertices(index_set)
        elif t is not None:
            self.add_vertices(t.index_set())

    def _repr_(self, compact=False):
        """
        EXAMPLES::

            sage: DynkinDiagram(['G',2])     # indirect doctest
              3
            O=<=O
            1   2
            G2
        """
        ct = self.cartan_type()
        result = ct.ascii_art() + "\n" if hasattr(ct, "ascii_art") else ""

        if ct is None or isinstance(ct, CartanMatrix):
            return result+"Dynkin diagram of rank %s" % self.rank()
        return result+"%s" % ct._repr_(compact=True)

    def _rich_repr_(self, display_manager, **kwds):
        """
        Rich Output Magic Method.

        Override rich output because :meth:`_repr_` outputs ascii
        art. The proper fix will be in :issue:`18328`.

        See :mod:`sage.repl.rich_output` for details.

        EXAMPLES::

            sage: from sage.repl.rich_output import get_display_manager
            sage: dm = get_display_manager()
            sage: E8 = WeylCharacterRing('E8')
            sage: E8.dynkin_diagram()._rich_repr_(dm)
            OutputAsciiArt container
        """
        OutputAsciiArt = display_manager.types.OutputAsciiArt
        OutputPlainText = display_manager.types.OutputPlainText
        if OutputAsciiArt in display_manager.supported_output():
            return OutputAsciiArt(self._repr_())
        return OutputPlainText(self._repr_())

    def _latex_(self, scale=0.5):
        r"""
        Return a latex representation of this Dynkin diagram.

        EXAMPLES::

            sage: latex(DynkinDiagram(['A',3,1]))
            \begin{tikzpicture}[scale=0.5]
            \draw (-1,0) node[anchor=east] {$A_{3}^{(1)}$};
            \draw (0 cm,0) -- (4 cm,0);
            \draw (0 cm,0) -- (2.0 cm, 1.2 cm);
            \draw (2.0 cm, 1.2 cm) -- (4 cm, 0);
            \draw[fill=white] (0 cm, 0 cm) circle (.25cm) node[below=4pt]{$1$};
            \draw[fill=white] (2 cm, 0 cm) circle (.25cm) node[below=4pt]{$2$};
            \draw[fill=white] (4 cm, 0 cm) circle (.25cm) node[below=4pt]{$3$};
            \draw[fill=white] (2.0 cm, 1.2 cm) circle (.25cm) node[anchor=south east]{$0$};
            \end{tikzpicture}
        """
        if self.cartan_type() is None:
            return "Dynkin diagram of rank {}".format(self.rank())

        from sage.graphs.graph_latex import setup_latex_preamble
        setup_latex_preamble()

        ret = "\\begin{{tikzpicture}}[scale={}]\n".format(scale)
        ret += "\\draw (-1,0) node[anchor=east] {{${}$}};\n".format(self.cartan_type()._latex_())
        ret += self.cartan_type()._latex_dynkin_diagram()
        ret += "\\end{tikzpicture}"
        return ret

    def _matrix_(self):
        """
        Return a regular matrix from ``self``.

        EXAMPLES::

            sage: M = DynkinDiagram(['C',3])._matrix_(); M
            [ 2 -1  0]
            [-1  2 -2]
            [ 0 -1  2]
            sage: type(M)
            <class 'sage.combinat.root_system.cartan_matrix.CartanMatrix'>
        """
        return self.cartan_matrix()._matrix_()

    def add_edge(self, i, j, label=1):
        """
        EXAMPLES::

            sage: from sage.combinat.root_system.dynkin_diagram import DynkinDiagram_class
            sage: d = DynkinDiagram_class(CartanType(['A',3]))
            sage: sorted(d.edges(sort=True))
            []
            sage: d.add_edge(2, 3)
            sage: sorted(d.edges(sort=True))
            [(2, 3, 1), (3, 2, 1)]
        """
        DiGraph.add_edge(self, i, j, label)
        if not self.has_edge(j,i):
            self.add_edge(j,i,1)

    def __hash__(self):
        """
        EXAMPLES::

            sage: d = CartanType(['A',3]).dynkin_diagram()
            sage: dv = d.vertices(sort=True)
            sage: hash(d) == hash((d.cartan_type(), tuple(dv), tuple(d.edge_iterator(dv))))
            True
        """
        # Should assert for immutability!

        #return hash(self.cartan_type(), self.vertices(sort=True), tuple(self.edges(sort=True)))
        # FIXME: self.edges() currently tests at some point whether
        # self is a vertex of itself which causes an infinite
        # recursion loop. Current workaround: call self.edge_iterator directly
        verts = self.vertices(sort=True)
        return hash((self.cartan_type(), tuple(verts), tuple(self.edge_iterator(verts))))

    @staticmethod
    def an_instance():
        """
        Return an example of Dynkin diagram.

        EXAMPLES::

            sage: from sage.combinat.root_system.dynkin_diagram import DynkinDiagram_class
            sage: g = DynkinDiagram_class.an_instance()
            sage: g
            Dynkin diagram of rank 3
            sage: g.cartan_matrix()
            [ 2 -1 -1]
            [-2  2 -1]
            [-1 -1  2]
        """
        # hyperbolic Dynkin diagram of Exercise 4.9 p. 57 of Kac Infinite Dimensional Lie Algebras.
        g = DynkinDiagram()
        g.add_vertices([1,2,3])
        g.add_edge(1,2,2)
        g.add_edge(1,3)
        g.add_edge(2,3)
        return g

    ##########################################################################
    # Cartan type methods

    @cached_method
    def index_set(self):
        """
        EXAMPLES::

            sage: DynkinDiagram(['C',3]).index_set()
            (1, 2, 3)
            sage: DynkinDiagram("A2","B2","F4").index_set()
            (1, 2, 3, 4, 5, 6, 7, 8)
        """
        return tuple(self.vertices(sort=True))

    def cartan_type(self):
        """
        EXAMPLES::

            sage: DynkinDiagram("A2","B2","F4").cartan_type()
            A2xB2xF4
        """
        return self._cartan_type

    def rank(self):
        r"""
        Return the index set for this Dynkin diagram.

        EXAMPLES::

            sage: DynkinDiagram(['C',3]).rank()
            3
            sage: DynkinDiagram("A2","B2","F4").rank()
            8
        """
        return self.n_vertices()

    def dynkin_diagram(self):
        """
        EXAMPLES::

            sage: DynkinDiagram(['C',3]).dynkin_diagram()
            O---O=<=O
            1   2   3
            C3
        """
        return self

    @cached_method
    def cartan_matrix(self):
        r"""
        Return the Cartan matrix for this Dynkin diagram.

        EXAMPLES::

            sage: DynkinDiagram(['C',3]).cartan_matrix()
            [ 2 -1  0]
            [-1  2 -2]
            [ 0 -1  2]
        """
        return CartanMatrix(self)

    def dual(self):
        r"""
        Return the dual Dynkin diagram, obtained by reversing all edges.

        EXAMPLES::

            sage: D = DynkinDiagram(['C',3])
            sage: D.edges(sort=True)
            [(1, 2, 1), (2, 1, 1), (2, 3, 1), (3, 2, 2)]
            sage: D.dual()
            O---O=>=O
            1   2   3
            B3
            sage: D.dual().edges(sort=True)
            [(1, 2, 1), (2, 1, 1), (2, 3, 2), (3, 2, 1)]
            sage: D.dual() == DynkinDiagram(['B',3])
            True

        TESTS::

            sage: D = DynkinDiagram(['A',0]); D
            A0
            sage: D.edges(sort=True)
            []
            sage: D.dual()
            A0
            sage: D.dual().edges(sort=True)
            []
            sage: D = DynkinDiagram(['A',1])
            sage: D.edges(sort=True)
            []
            sage: D.dual()
            O
            1
            A1
            sage: D.dual().edges(sort=True)
            []
        """
        result = DynkinDiagram_class(None, odd_isotropic_roots=self._odd_isotropic_roots)
        result.add_vertices(self.vertices(sort=True))
        for source, target, label in self.edges(sort=False):
            result.add_edge(target, source, label)
        result._cartan_type = self._cartan_type.dual() if self._cartan_type is not None else None
        return result

    def relabel(self, *args, **kwds):
        """
        Return the relabelled Dynkin diagram of ``self``.

        INPUT: see :meth:`~sage.graphs.generic_graph.GenericGraph.relabel`

        There is one difference: the default value for ``inplace`` is
        ``False`` instead of ``True``.

        EXAMPLES::

            sage: D = DynkinDiagram(['C',3])
            sage: D.relabel({1:0, 2:4, 3:1})
            O---O=<=O
            0   4   1
            C3 relabelled by {1: 0, 2: 4, 3: 1}
            sage: D
            O---O=<=O
            1   2   3
            C3

            sage: _ = D.relabel({1:0, 2:4, 3:1}, inplace=True)
            sage: D
            O---O=<=O
            0   4   1
            C3 relabelled by {1: 0, 2: 4, 3: 1}

            sage: D = DynkinDiagram(['A', [1,2]])
            sage: Dp = D.relabel({-1:4, 0:-3, 1:3, 2:2})
            sage: Dp
            O---X---O---O
            4   -3  3   2
            A1|2 relabelled by {-1: 4, 0: -3, 1: 3, 2: 2}
            sage: Dp.odd_isotropic_roots()
            (-3,)

            sage: D = DynkinDiagram(['D', 5])
            sage: G, perm = D.relabel(range(5), return_map=True)
            sage: G
                    O 4
                    |
                    |
            O---O---O---O
            0   1   2   3
            D5 relabelled by {1: 0, 2: 1, 3: 2, 4: 3, 5: 4}
            sage: perm
            {1: 0, 2: 1, 3: 2, 4: 3, 5: 4}

            sage: perm = D.relabel(range(5), return_map=True, inplace=True)
            sage: D
                    O 4
                    |
                    |
            O---O---O---O
            0   1   2   3
            D5 relabelled by {1: 0, 2: 1, 3: 2, 4: 3, 5: 4}
            sage: perm
            {1: 0, 2: 1, 3: 2, 4: 3, 5: 4}
        """
        return_map = kwds.pop("return_map", False)
        inplace = kwds.pop("inplace", False)
        if inplace:
            G = self
        else:
            # We need to copy self because we want to return the
            # permutation and that works when relabelling in place.
            G = self.copy()

        perm = DiGraph.relabel(G, *args, inplace=True, return_map=True, **kwds)
        new_odds = [perm[i] for i in self._odd_isotropic_roots]
        G._odd_isotropic_roots = tuple(new_odds)
        if self._cartan_type is not None:
            G._cartan_type = self._cartan_type.relabel(perm.__getitem__)
        if return_map:
            if inplace:
                return perm
            return G, perm
        return G

    def subtype(self, index_set):
        """
        Return a subtype of ``self`` given by ``index_set``.

        A subtype can be considered the Dynkin diagram induced from
        the Dynkin diagram of ``self`` by ``index_set``.

        EXAMPLES::

            sage: D = DynkinDiagram(['A',6,2]); D
            O=<=O---O=<=O
            0   1   2   3
            BC3~
            sage: D.subtype([1,2,3])
            Dynkin diagram of rank 3
        """
        return self.cartan_matrix().subtype(index_set).dynkin_diagram()

    def is_finite(self) -> bool:
        """
        Check if ``self`` corresponds to a finite root system.

        EXAMPLES::

            sage: CartanType(['F',4]).dynkin_diagram().is_finite()
            True
            sage: D = DynkinDiagram(CartanMatrix([[2, -4], [-3, 2]]))
            sage: D.is_finite()
            False
        """
        if self._cartan_type is not None:
            return self._cartan_type.is_finite()
        return self.cartan_matrix().is_finite()

    def is_affine(self) -> bool:
        """
        Check if ``self`` corresponds to an affine root system.

        EXAMPLES::

            sage: CartanType(['F',4]).dynkin_diagram().is_affine()
            False
            sage: D = DynkinDiagram(CartanMatrix([[2, -4], [-3, 2]]))
            sage: D.is_affine()
            False
        """
        if self._cartan_type is not None:
            return self._cartan_type.is_affine()
        return self.cartan_matrix().is_affine()

    def is_irreducible(self) -> bool:
        """
        Check if ``self`` corresponds to an irreducible root system.

        EXAMPLES::

            sage: CartanType(['F',4]).dynkin_diagram().is_irreducible()
            True
            sage: CM = CartanMatrix([[2,-6],[-4,2]])
            sage: CM.dynkin_diagram().is_irreducible()
            True
            sage: CartanType("A2xB3").dynkin_diagram().is_irreducible()
            False
            sage: CM = CartanMatrix([[2,-6,0],[-4,2,0],[0,0,2]])
            sage: CM.dynkin_diagram().is_irreducible()
            False
        """
        if self._cartan_type is not None:
            return self._cartan_type.is_irreducible()
        return self.number_of_connected_components() == 1

    def is_crystallographic(self) -> bool:
        """
        Implement :meth:`CartanType_abstract.is_crystallographic`.

        A Dynkin diagram always corresponds to a crystallographic root system.

        EXAMPLES::

            sage: CartanType(['F',4]).dynkin_diagram().is_crystallographic()
            True

        TESTS::

            sage: CartanType(['G',2]).dynkin_diagram().is_crystallographic()
            True
        """
        return True

    def symmetrizer(self):
        """
        Return the symmetrizer of the corresponding Cartan matrix.

        EXAMPLES::

            sage: d = DynkinDiagram()
            sage: d.add_edge(1,2,3)
            sage: d.add_edge(2,3)
            sage: d.add_edge(3,4,3)
            sage: d.symmetrizer()
            Finite family {1: 9, 2: 3, 3: 3, 4: 1}

        TESTS:

        We check that :issue:`15740` is fixed::

            sage: d = DynkinDiagram()
            sage: d.add_edge(1,2,3)
            sage: d.add_edge(2,3)
            sage: d.add_edge(3,4,3)
            sage: L = d.root_system().root_lattice()
            sage: al = L.simple_roots()
            sage: al[1].associated_coroot()
            alphacheck[1]
            sage: al[1].reflection(al[2])
            alpha[1] + 3*alpha[2]
        """
        return self.cartan_matrix().symmetrizer()

    def odd_isotropic_roots(self):
        """
        Return the odd isotropic roots of ``self``.

        EXAMPLES::

            sage: g = DynkinDiagram(['A',4])
            sage: g.odd_isotropic_roots()
            ()
            sage: g = DynkinDiagram(['A',[4,3]])
            sage: g.odd_isotropic_roots()
            (0,)
        """
        return self._odd_isotropic_roots

    def __getitem__(self, i):
        r"""
        With a tuple (i,j) as argument, returns the scalar product
        `\langle \alpha^\vee_i, \alpha_j\rangle`.

        Otherwise, behaves as the usual ``DiGraph.__getitem__``

        EXAMPLES:

        We use the `C_4` Dynkin diagram as a Cartan matrix::

            sage: g = DynkinDiagram(['C',4])
            sage: matrix([[g[i,j] for j in range(1,5)] for i in range(1,5)])
            [ 2 -1  0  0]
            [-1  2 -1  0]
            [ 0 -1  2 -2]
            [ 0  0 -1  2]

        The neighbors of a node can still be obtained in the usual way::

            sage: [g[i] for i in range(1,5)]
            [[2], [1, 3], [2, 4], [3]]
        """
        if not isinstance(i, tuple):
            return DiGraph.__getitem__(self, i)
        i, j = i
        if i == j:
            if i in self._odd_isotropic_roots:
                return 0
            return 2
        if self.has_edge(j, i):
            return -self.edge_label(j, i)
        return 0

    def column(self, j):
        """
        Return the `j`-th column `(a_{i,j})_i` of the
        Cartan matrix corresponding to this Dynkin diagram, as a container
        (or iterator) of tuples `(i, a_{i,j})`.

        EXAMPLES::

            sage: g = DynkinDiagram(["B",4])
            sage: [ (i,a) for (i,a) in g.column(3) ]
            [(3, 2), (2, -1), (4, -2)]
        """
        val = 2 if j not in self._odd_isotropic_roots else 0
        return [(j,val)] + [(i,-m) for (j1, i, m) in self.outgoing_edges(j)]

    def row(self, i):
        """
        Return the `i`-th row `(a_{i,j})_j` of the
        Cartan matrix corresponding to this Dynkin diagram, as a container
        (or iterator) of tuples `(j, a_{i,j})`.

        EXAMPLES::

            sage: g = DynkinDiagram(["C",4])
            sage: [ (i,a) for (i,a) in g.row(3) ]
            [(3, 2), (2, -1), (4, -2)]
        """
        val = 2 if i not in self._odd_isotropic_roots else 0
        return [(i,val)] + [(j,-m) for (j, i1, m) in self.incoming_edges(i)]

    @cached_method
    def coxeter_diagram(self):
        r"""
        Construct the Coxeter diagram of ``self``.

        .. SEEALSO:: :meth:`CartanType_abstract.coxeter_diagram`

        EXAMPLES::

            sage: cm = CartanMatrix([[2,-5,0],[-2,2,-1],[0,-1,2]])
            sage: D = cm.dynkin_diagram()
            sage: G = D.coxeter_diagram(); G
            Graph on 3 vertices
            sage: G.edges(sort=True)
            [(0, 1, +Infinity), (1, 2, 3)]

            sage: ct = CartanType([['A',2,2], ['B',3]])
            sage: ct.coxeter_diagram()
            Graph on 5 vertices
            sage: ct.dynkin_diagram().coxeter_diagram() == ct.coxeter_diagram()
            True
        """
        from sage.rings.infinity import infinity
        scalarproducts_to_order = {0: 2, 1: 3, 2: 4, 3: 6}
        from sage.graphs.graph import Graph
        coxeter_diagram = Graph(multiedges=False)
        I = self.index_set()
        coxeter_diagram.add_vertices(I)
        for i in I:
            for j in self.neighbors_out(i):
                # avoid adding the edge twice
                if not coxeter_diagram.has_edge(i,j):
                    val = scalarproducts_to_order.get(self[i,j]*self[j,i], infinity)
                    coxeter_diagram.add_edge(i,j, val)
        return coxeter_diagram.copy(immutable=True)


def precheck(t, letter=None, length=None, affine=None, n_ge=None, n=None):
    """
    EXAMPLES::

        sage: from sage.combinat.root_system.dynkin_diagram import precheck
        sage: ct = CartanType(['A',4])
        sage: precheck(ct, letter='C')
        Traceback (most recent call last):
        ...
        ValueError: t[0] must be = 'C'
        sage: precheck(ct, affine=1)
        Traceback (most recent call last):
        ...
        ValueError: t[2] must be = 1
        sage: precheck(ct, length=3)
        Traceback (most recent call last):
        ...
        ValueError: len(t) must be = 3
        sage: precheck(ct, n=3)
        Traceback (most recent call last):
        ...
        ValueError: t[1] must be = 3
        sage: precheck(ct, n_ge=5)
        Traceback (most recent call last):
        ...
        ValueError: t[1] must be >= 5
    """
    if letter is not None:
        if t[0] != letter:
            raise ValueError("t[0] must be = '%s'" % letter)

    if length is not None:
        if len(t) != length:
            raise ValueError("len(t) must be = %s" % length)

    if affine is not None:
        try:
            if t[2] != affine:
                raise ValueError("t[2] must be = %s" % affine)
        except IndexError:
            raise ValueError("t[2] must be = %s" % affine)

    if n_ge is not None:
        if t[1] < n_ge:
            raise ValueError("t[1] must be >= %s" % n_ge)

    if n is not None:
        if t[1] != n:
            raise ValueError("t[1] must be = %s" % n)



"""
Common globals defined by GAP.
"""

###############################################################################
#       Copyright (C) 2009, William Stein <wstein@gmail.com>
#       Copyright (C) 2012, Volker Braun <vbraun.name@gmail.com>
#
#   Distributed under the terms of the GNU General Public License (GPL)
#   as published by the Free Software Foundation; either version 2 of
#   the License, or (at your option) any later version.
#                   https://www.gnu.org/licenses/
###############################################################################


from .gap_functions import common_gap_functions


# selected gap globals to use in tab completion
common_gap_globals = {
    'Assert',
    'Cyclotomics',
    'GaussianIntegers',
    'GaussianRationals',
    'GlobalMersenneTwister',
    'GlobalRandomSource',
    'InfoAlgebra',
    'InfoAttributes',
    'InfoBckt',
    'InfoCharacterTable',
    'InfoCoh',
    'InfoComplement',
    'InfoCoset',
    'InfoFpGroup',
    'InfoGroebner',
    'InfoGroup',
    'InfoLattice',
    'InfoMatrix',
    'InfoMonomial',
    'InfoNumtheor',
    'InfoOptions',
    'InfoPackageLoading',
    'InfoPcSubgroup',
    'InfoWarning',
    'Integers',
    'NiceBasisFiltersInfo',
    'Primes',
    'Rationals',
    'TableOfMarksComponents'
} | common_gap_functions
